import numpy as np
import scipy.sparse as sp
from scipy.optimize import linprog
from scipy.spatial import HalfspaceIntersection
from cached_property import cached_property

from stochastic_matching.common import (
    pseudo_inverse_scalar,
    clean_zeros,
    CharMaker,
    neighbors,
    class_converter,
)
from stochastic_matching.display import show
from stochastic_matching.simulator.simulator import Simulator

status_names = {
    (False, False): "Nonjective",
    (True, False): "Injective-only",
    (False, True): "Surjective-only",
    (True, True): "Bijective",
}
"""Name associated to a (injective, surjective) tuple."""

status_names_simple_connected = {
    (False, False): "Bipartite with cycle(s)",
    (True, False): "Tree",
    (False, True): "Non-bipartite polycyclic",
    (True, True): "Non-bipartite monocyclic",
}
"""Name associated to a (injective, surjective) tuple when the graph is simple and connected."""


def adjacency_to_incidence(adjacency):
    """
    Converts adjacency matrix to incidence matrix.

    Parameters
    ----------
    adjacency: :class:`~numpy.ndarray`
        Adjacency matrix of a simple graph (symmetric matrix with 0s and 1s, null diagonal).

    Returns
    -------
    :class:`~numpy.ndarray`
        Incidence matrix between nodes and edges.

    Examples
    --------

    Convert a diamond adjacency to incidence.

    >>> diamond = np.array([[0, 1, 1, 0],
    ...           [1, 0, 1, 1],
    ...           [1, 1, 0, 1],
    ...           [0, 1, 1, 0]])
    >>> adjacency_to_incidence(diamond)
    array([[1, 1, 0, 0, 0],
           [1, 0, 1, 1, 0],
           [0, 1, 1, 0, 1],
           [0, 0, 0, 1, 1]])
    """
    n, _ = adjacency.shape
    edges = [(i, j) for i in range(n) for j in range(i + 1, n) if adjacency[i, j]]
    m = len(edges)
    incidence = np.zeros((n, m), dtype=int)
    for j, e in enumerate(edges):
        for i in e:
            incidence[i, j] = 1
    return incidence


def incidence_to_adjacency(incidence):
    """
    Converts incidence matrix to adjacency matrix.
    If the incidence matrix does not correspond to a simple graph, an error is thrown.

    Parameters
    ----------
    incidence: :class:`~numpy.ndarray`
        Incidence matrix of a simple graph (matrix with 0s and 1s, two 1s per column).

    Returns
    -------
    :class:`~numpy.ndarray`
        Adjacency matrix.

    Examples
    --------

    Convert a diamond graph from incidence to adjacency.

    >>> import stochastic_matching as sm
    >>> diamond = sm.CycleChain()
    >>> diamond.incidence
    array([[1, 1, 0, 0, 0],
           [1, 0, 1, 1, 0],
           [0, 1, 1, 0, 1],
           [0, 0, 0, 1, 1]])
    >>> incidence_to_adjacency(diamond.incidence)
    array([[0, 1, 1, 0],
           [1, 0, 1, 1],
           [1, 1, 0, 1],
           [0, 1, 1, 0]])

    An error occurs if one tries to convert a hypergraph.

    >>> candy = sm.HyperPaddle()
    >>> candy.incidence
    array([[1, 1, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0],
           [0, 1, 1, 0, 0, 0, 1],
           [0, 0, 0, 0, 0, 0, 1],
           [0, 0, 0, 1, 1, 0, 1],
           [0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 1, 1, 0]])
    >>> incidence_to_adjacency(candy.incidence)
    Traceback (most recent call last):
    ...
    ValueError: The incidence matrix does not seem to correspond to a simple graph.
    """
    # noinspection PyUnresolvedReferences
    if not np.all(np.sum(incidence, axis=0) == 2):
        raise ValueError(
            "The incidence matrix does not seem to correspond to a simple graph."
        )
    incidence = sp.csc_matrix(incidence)
    n, m = incidence.shape
    adjacency = np.zeros((n, n), dtype=int)
    for j in range(m):
        e = neighbors(j, incidence)
        adjacency[e[0], e[1]] = 1
        adjacency[e[1], e[0]] = 1
    return adjacency


class Kernel:
    """
    Parameters
    ----------
    incidence: :class:`~numpy.ndarray`
        Incidence matrix of the graph to analyze.
    tol: :class:`float`
        Tolerance for approximating zero.

    Attributes
    -----------

    Examples
    --------

    >>> import stochastic_matching.graphs as sm
    >>> paw = sm.Tadpole()
    >>> kernel  = Kernel(paw.incidence)

    The inverse is:

    >>> kernel.inverse
    array([[ 0.5,  0.5, -0.5,  0.5],
           [ 0.5, -0.5,  0.5, -0.5],
           [-0.5,  0.5,  0.5, -0.5],
           [ 0. ,  0. ,  0. ,  1. ]])

    We can check that it is indeed the inverse.

    >>> i = paw.incidence @ kernel.inverse
    >>> clean_zeros(i)
    >>> i
    array([[1., 0., 0., 0.],
           [0., 1., 0., 0.],
           [0., 0., 1., 0.],
           [0., 0., 0., 1.]])

    Right kernel is trivial:

    >>> kernel.right.shape[0]
    0

    Left kernel is trivial:

    >>> kernel.left.shape[1]
    0

    Graph is bijective:

    >>> kernel.type
    'Bijective'

    As the graph is simple and connected, there a more accurate description of the status:

    >>> status_names_simple_connected[kernel.status]
    'Non-bipartite monocyclic'

    A summary:

    >>> kernel
    Kernels of a graph with 4 nodes and 4 edges.
    Node dimension is 0.
    Edge dimension is 0
    Type: Bijective
    Node kernel:
    []
    Edge kernel:
    []

    Now consider a bipartite version, the banner graph :

    >>> banner = sm.Tadpole(m=4)
    >>> kernel = Kernel(banner.incidence)

    The pseudo-inverse is:

    >>> kernel.inverse
    array([[ 0.35,  0.4 , -0.15, -0.1 ,  0.1 ],
           [ 0.45, -0.2 , -0.05,  0.3 , -0.3 ],
           [-0.15,  0.4 ,  0.35, -0.1 ,  0.1 ],
           [-0.05, -0.2 ,  0.45,  0.3 , -0.3 ],
           [-0.2 ,  0.2 , -0.2 ,  0.2 ,  0.8 ]])

    We can check that it is indeed not exactly the inverse.

    >>> i = banner.incidence @ kernel.inverse
    >>> i
    array([[ 0.8,  0.2, -0.2,  0.2, -0.2],
           [ 0.2,  0.8,  0.2, -0.2,  0.2],
           [-0.2,  0.2,  0.8,  0.2, -0.2],
           [ 0.2, -0.2,  0.2,  0.8,  0.2],
           [-0.2,  0.2, -0.2,  0.2,  0.8]])

    Right kernel is not trivial because of the even cycle:

    >>> kernel.right.shape[0]
    1
    >>> kernel.right # doctest: +SKIP
    array([[ 0.5, -0.5, -0.5,  0.5,  0. ]])

    Left kernel is not trivial because of the bipartite degenerescence:

    >>> kernel.left.shape[1]
    1
    >>> kernel.left
    array([[ 0.4472136],
           [-0.4472136],
           [ 0.4472136],
           [-0.4472136],
           [ 0.4472136]])

    Status is nonjective (not injective nor bijective):

    >>> kernel.type
    'Nonjective'

    As the graph is simple and connected, there a more accurate description of the status:

    >>> status_names_simple_connected[kernel.status]
    'Bipartite with cycle(s)'

    Consider now the diamond graph, surjective (n<m, non bipartite).

    >>> diamond = sm.CycleChain()
    >>> kernel = Kernel(diamond.incidence)

    The inverse is:

    >>> kernel.inverse
    array([[ 0.5 ,  0.25, -0.25,  0.  ],
           [ 0.5 , -0.25,  0.25,  0.  ],
           [-0.5 ,  0.5 ,  0.5 , -0.5 ],
           [ 0.  ,  0.25, -0.25,  0.5 ],
           [ 0.  , -0.25,  0.25,  0.5 ]])

    We can check that it is indeed the inverse.

    >>> i = diamond.incidence @ kernel.inverse
    >>> clean_zeros(i)
    >>> i
    array([[1., 0., 0., 0.],
           [0., 1., 0., 0.],
           [0., 0., 1., 0.],
           [0., 0., 0., 1.]])

    There is a right kernel:

    >>> kernel.right.shape[0]
    1
    >>> kernel.right
    array([[ 0.5, -0.5,  0. , -0.5,  0.5]])

    Normalized version:

    >>> kernel.alt_right
    array([[ 1., -1.,  0., -1.,  1.]])

    The left kernel is trivial:

    >>> kernel.left.shape[1]
    0

    The diamond is surjective-only:

    >>> kernel.type
    'Surjective-only'

    As the graph is simple and connected, there a more accurate description of the status:

    >>> status_names_simple_connected[kernel.status]
    'Non-bipartite polycyclic'

    Consider now a star graph, injective (tree).

    >>> star = sm.Star()
    >>> kernel = Kernel(star.incidence)

    The inverse is:

    >>> kernel.inverse
    array([[ 0.25,  0.75, -0.25, -0.25],
           [ 0.25, -0.25,  0.75, -0.25],
           [ 0.25, -0.25, -0.25,  0.75]])

    We can check that it is indeed the **left** inverse.

    >>> i = kernel.inverse @ star.incidence
    >>> clean_zeros(i)
    >>> i
    array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]])

    The right kernel is trivial:

    >>> kernel.right.shape[0]
    0

    The left kernel shows the bibartite behavior:

    >>> kernel.left.shape[1]
    1
    >>> kernel.left
    array([[-0.5],
           [ 0.5],
           [ 0.5],
           [ 0.5]])

    The star is injective-only:

    >>> kernel.type
    'Injective-only'

    As the graph is simple and connected, there a more accurate description of the status:

    >>> status_names_simple_connected[kernel.status]
    'Tree'

    Next, a surjective hypergraph:

    >>> clover = sm.Fan()
    >>> kernel = Kernel(clover.incidence)

    Incidence matrix dimensions:

    >>> clover.incidence.shape
    (9, 10)

    The inverse dimensions:

    >>> kernel.inverse.shape
    (10, 9)

    We can check that it is exactly the inverse, because there was no dimensionnality loss.

    >>> i = clover.incidence @ kernel.inverse
    >>> clean_zeros(i)
    >>> i
    array([[1., 0., 0., 0., 0., 0., 0., 0., 0.],
           [0., 1., 0., 0., 0., 0., 0., 0., 0.],
           [0., 0., 1., 0., 0., 0., 0., 0., 0.],
           [0., 0., 0., 1., 0., 0., 0., 0., 0.],
           [0., 0., 0., 0., 1., 0., 0., 0., 0.],
           [0., 0., 0., 0., 0., 1., 0., 0., 0.],
           [0., 0., 0., 0., 0., 0., 1., 0., 0.],
           [0., 0., 0., 0., 0., 0., 0., 1., 0.],
           [0., 0., 0., 0., 0., 0., 0., 0., 1.]])

    Right kernel is 1 dimensional:

    >>> kernel.right.shape[0]
    1

    Left kernel is trivial.

    >>> kernel.left.shape[1]
    0

    Status:

    >>> kernel.type
    'Surjective-only'

    Lastly, observe a *bipartite* hypergraph (in the sense of with non-trivial left kernel).

    >>> clover = sm.Fan(cycle_size=4)
    >>> kernel = Kernel(clover.incidence)

    Incidence matrix dimensions:

    >>> clover.incidence.shape
    (12, 13)

    The inverse dimensions:

    >>> kernel.inverse.shape
    (13, 12)

    We can check that it is not exactly the inverse.

    >>> (clover.incidence @ kernel.inverse)[:4, :4]
    array([[ 0.83333333,  0.16666667, -0.16666667,  0.16666667],
           [ 0.16666667,  0.83333333,  0.16666667, -0.16666667],
           [-0.16666667,  0.16666667,  0.83333333,  0.16666667],
           [ 0.16666667, -0.16666667,  0.16666667,  0.83333333]])

    Right kernel is 3 dimensional:

    >>> kernel.right.shape[0]
    3

    Left kernel is 2-dimensional (this is a change compared to simple graph,
    where the left kernel dimension of a connected component is at most 1).

    >>> kernel.left.shape[1]
    2

    Status:

    >>> kernel.type
    'Nonjective'
    """

    def __init__(self, incidence, tol=1e-10):
        self.tol = tol
        self.right = None
        self.left = None
        self.inverse = None
        self.right_inverse = None
        self.status = None
        self.type = None
        self.fit(incidence)

    @property
    def alt_right(self):
        live_edges = self.right[self.right !=0]
        if len(live_edges) > 0:
            return self.right / np.min(np.abs(live_edges))
        else:
            return self.right

    def fit(self, incidence):
        n, m = incidence.shape
        min_d = min(n, m)
        u, s, v = np.linalg.svd(incidence)
        clean_zeros(s, tol=self.tol)
        dia = np.zeros((m, n))
        dia[:min_d, :min_d] = np.diag([pseudo_inverse_scalar(e) for e in s])
        ev = np.zeros(m)
        ev[: len(s)] = s
        self.right = v[ev == 0, :]
        eu = np.zeros(n)
        eu[: len(s)] = s
        self.left = u[:, eu == 0]
        self.inverse = v.T @ dia @ u.T
        self.right_inverse = kernel_inverse(self.right)
        clean_zeros(self.inverse, tol=self.tol)
        clean_zeros(self.right, tol=self.tol)
        clean_zeros(self.left, tol=self.tol)
        clean_zeros(self.right_inverse, tol=self.tol)
        injective = self.right.shape[0] == 0
        surjective = self.left.shape[1] == 0
        self.status = (injective, surjective)
        self.type = status_names[self.status]

    def __repr__(self):
        return (
            f"Kernels of a graph with {self.left.shape[0]} nodes and {self.right.shape[1]} edges.\n"
            f"Node dimension is {self.left.shape[1]}.\n"
            f"Edge dimension is {self.right.shape[0]}\n"
            f"Type: {self.type}\n"
            f"Node kernel:\n"
            f"{self.left}\n"
            f"Edge kernel:\n"
            f"{self.right}"
        )


def kernel_inverse(kernel):
    """
    Parameters
    ----------
    kernel: :class:`numpy.ndarray`
        Matrix of kernel vectors (not necessarily orthogonal) of shape dXm.

    Returns
    -------
    :class:`numpy.ndarray`
        The `reverse` matrix dXd that allows to transform inner product with kernel to kernel coordinates.

    Examples
    --------

    When the kernel basis is orthogonal,
    it returns the diagonal matrix with the inverse of the squared norm of the vectors.
    For example:

    >>> edge_kernel = np.array([[ 0,  0,  1, -1, -1,  1,  0,  0],
    ...       [ 1, -1,  0, -1,  1,  0, -1,  1]])
    >>> inverse = kernel_inverse(edge_kernel)
    >>> inverse
    array([[0.25      , 0.        ],
           [0.        , 0.16666667]])

    Here the kernel basis is orthogonal,
    so you just have a diagonal matrix with the inverse square of the vector norms.

    Consider [-1, 3] in kernel coordinates. In edge coordinates, it corresponds to:

    >>> edges = np.array([-1, 3]) @ edge_kernel
    >>> edges
    array([ 3, -3, -1, -2,  4, -1, -3,  3])

    One can project the edges on the kernel.

    >>> projection = edge_kernel @ edges
    >>> projection
    array([-4, 18])

    The kernel inverse allows to rectify the projection in kernel coordinates.

    >>> projection @ inverse
    array([-1.,  3.])

    If the kernel basis is not orthogonal, it returns somethings more complex.

    >>> edge_kernel = np.array([[ 1, -1,  1, -2,  0,  1, -1,  1],
    ...    [ 0,  0, -1,  1,  1, -1,  0,  0]])
    >>> inverse = kernel_inverse(edge_kernel)
    >>> inverse
    array([[0.16666667, 0.16666667],
           [0.16666667, 0.41666667]])

    The inverse is more complex because the basis is not orthogonal.

    >>> edges = np.array([2, -1]) @ edge_kernel
    >>> edges
    array([ 2, -2,  3, -5, -1,  3, -2,  2])

    >>> projection = edge_kernel @ edges
    >>> projection
    array([ 24, -12])

    >>> projection @ inverse
    array([ 2., -1.])
    """
    return np.linalg.inv(np.inner(kernel, kernel))


def simple_right_kernel(edge_kernel, seeds):
    """
    Parameters
    ----------
    edge_kernel: :class:`~numpy.ndarray`
        Right kernel (i.e. edges kernel) of a simple graph.
    seeds: :class:`list` of :class:`int`
        Seed edges of the kernel space. Valid seeds can be obtained from
        :meth:`~stochastic_matching.analysis.connected_components`.
        Cf https://hal.archives-ouvertes.fr/hal-03502084.

    Returns
    -------
    :class:`~numpy.ndarray`
        The kernel expressed as elements from the cycle space (even cycles and kayak paddles).

    Examples
    --------

    Start with the co-domino.

    >>> import stochastic_matching.graphs as sm

    Default decomposition is a square and an hex.

    >>> right = sm.Codomino().kernel.right
    >>> right
    array([[ 0,  0,  1, -1, -1,  1,  0,  0],
           [ 1, -1,  0, -1,  1,  0, -1,  1]])

    A second possible decomposition with a kayak paddle and a square.

    >>> simple_right_kernel(right, [0, 4])
    array([[ 1, -1,  1, -2,  0,  1, -1,  1],
           [ 0,  0, -1,  1,  1, -1,  0,  0]])

    Another example with the pyramid.

    >>> right = sm.Pyramid().kernel.right

    Default decomposition: cycles of length 8, 10, and 6.

    >>> right
    array([[ 0,  0,  1, -1, -1,  1,  0, -1,  1, -1,  1,  0,  0],
           [-1,  1,  0,  1, -1,  1,  0, -1, -1,  1,  0, -1,  1],
           [ 1, -1,  0, -1,  1, -1,  1,  0,  0,  0,  0,  0,  0]])

    Second decomposition: two cycles of length 6, one of length 8.

    >>> simple_right_kernel(right, [0, 12, 2])
    array([[ 1, -1,  0, -1,  1, -1,  1,  0,  0,  0,  0,  0,  0],
           [ 0,  0,  0,  0,  0,  0,  1, -1, -1,  1,  0, -1,  1],
           [ 0,  0,  1, -1, -1,  1,  0, -1,  1, -1,  1,  0,  0]])

    Another decomposition: two cycles of length 6 and a kayak paddle :math:`KP_{3, 3, 3}`.

    >>> simple_right_kernel(right, [5, 7, 2])
    array([[-1,  1,  0,  1, -1,  1, -1,  0,  0,  0,  0,  0,  0],
           [ 0,  0,  0,  0,  0,  0, -1,  1,  1, -1,  0,  1, -1],
           [ 1, -1,  1, -2,  0,  0,  0,  0,  2, -2,  1,  1, -1]])
    """
    return np.round(np.linalg.inv(edge_kernel[:, seeds]) @ edge_kernel).astype(int)


def simple_left_kernel(left):
    """

    Parameters
    ----------
    left: :class:`~numpy.ndarray`
        Left kernel (i.e. nodes kernel) of a simple graph, corresponding to bipartite components

    Returns
    -------
    :class:`~numpy.ndarray`
        The kernel with infinite-norm renormalization.

    Examples
    --------

    By default the kernel vector are 2-normalized.

    >>> import stochastic_matching.graphs as sm
    >>> sample = sm.concatenate([sm.Cycle(4), sm.Star(5)], 0)
    >>> raw_kernel = Kernel(sample.incidence)
    >>> raw_kernel.left
    array([[ 0.5      ,  0.       ],
           [-0.5      ,  0.       ],
           [ 0.5      ,  0.       ],
           [-0.5      ,  0.       ],
           [ 0.       , -0.4472136],
           [ 0.       ,  0.4472136],
           [ 0.       ,  0.4472136],
           [ 0.       ,  0.4472136],
           [ 0.       ,  0.4472136]])

    `simple_left_kernel` adjusts the values to {-1, 0, 1}

    >>> simple_left_kernel(raw_kernel.left)
    array([[ 1,  0],
           [-1,  0],
           [ 1,  0],
           [-1,  0],
           [ 0, -1],
           [ 0,  1],
           [ 0,  1],
           [ 0,  1],
           [ 0,  1]])
    """
    return np.around(left / np.max(left[:], axis=0)).astype(int)


def __intify_set(int_set):
    return {int(i) for i in int_set}


def traversal(model):
    """
    Using graph traversal, splits the graph into its connected components as a list of sets of nodes and edges.
    If the graph is simple, additional information obtained by the traversal are provided.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model with graph to decompose.

    Returns
    -------
    :class:`list` of :class:`dict`
        The list of connected components.
        If the graph is not simple, each connected component contains its sets of nodes and edges.
        If the graph is simple, each component also contains a set of spanning edges, a pivot edge that
        makes the spanner bijective (if any), a set set of edges that can seed the kernel space, and the type
        of the connected component.

    Examples
    ---------

    For simple graphs, the method provides a lot of information on each connected component.

    >>> import stochastic_matching.graphs as sm
    >>> tutti = sm.concatenate([sm.Cycle(4), sm.Complete(4), sm.CycleChain(), sm.Tadpole(), sm.Star()], 0)
    >>> traversal(tutti) # doctest: +NORMALIZE_WHITESPACE
    [{'nodes': {0, 1, 2, 3}, 'edges': {0, 1, 2, 3},
    'spanner': {0, 1, 2}, 'pivot': False, 'seeds': {3},
    'type': 'Bipartite with cycle(s)'},
    {'nodes': {4, 5, 6, 7}, 'edges': {4, 5, 6, 7, 8, 9},
    'spanner': {4, 5, 6}, 'pivot': 8, 'seeds': {9, 7},
    'type': 'Non-bipartite polycyclic'},
    {'nodes': {8, 9, 10, 11}, 'edges': {10, 11, 12, 13, 14},
    'spanner': {10, 11, 13}, 'pivot': 12, 'seeds': {14},
    'type': 'Non-bipartite polycyclic'},
    {'nodes': {12, 13, 14, 15}, 'edges': {16, 17, 18, 15},
    'spanner': {16, 18, 15}, 'pivot': 17, 'seeds': set(),
    'type': 'Non-bipartite monocyclic'},
    {'nodes': {16, 17, 18, 19}, 'edges': {19, 20, 21},
    'spanner': {19, 20, 21}, 'pivot': False, 'seeds': set(),
    'type': 'Tree'}]

    This information makes the analysis worthy even in the cases where the graph is connected.

    >>> traversal(sm.Pyramid()) # doctest: +NORMALIZE_WHITESPACE
    [{'nodes': {0, 1, 2, 3, 4, 5, 6, 7, 8, 9},
    'edges': {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12},
    'spanner': {0, 1, 3, 4, 5, 7, 8, 9, 11},
    'pivot': 2, 'seeds': {10, 12, 6},
    'type': 'Non-bipartite polycyclic'}]

    If the graph is treated as hypergraph, a lot less information is available.

    >>> tutti.adjacency = None
    >>> traversal(tutti) # doctest: +NORMALIZE_WHITESPACE
    [{'nodes': {0, 1, 2, 3}, 'edges': {0, 1, 2, 3}},
    {'nodes': {4, 5, 6, 7}, 'edges': {4, 5, 6, 7, 8, 9}},
    {'nodes': {8, 9, 10, 11}, 'edges': {10, 11, 12, 13, 14}},
    {'nodes': {12, 13, 14, 15}, 'edges': {16, 17, 18, 15}},
    {'nodes': {16, 17, 18, 19}, 'edges': {19, 20, 21}}]
    """
    simple = model.adjacency is not None
    n, m = model.n, model.m
    unknown_nodes = {i for i in range(n)}
    if simple:
        spin = np.ones(n, dtype=bool)  # Simple
    res = []
    while unknown_nodes:
        buffer = {unknown_nodes.pop()}
        current_nodes = set()
        current_edges = set()
        if simple:
            current_spanner = set()
        while buffer:
            i = buffer.pop()
            current_nodes.add(i)
            edges = neighbors(i, model.incidence_csr)
            for edge in edges:
                if edge in current_edges:
                    continue
                for j in neighbors(edge, model.incidence_csc):
                    if j in unknown_nodes:
                        buffer.add(j)
                        unknown_nodes.remove(j)
                        if simple:
                            spin[j] = not spin[i]  # Simple
                            current_spanner.add(edge)
                current_edges.add(edge)
        cc = {
            "nodes": __intify_set(current_nodes),
            "edges": __intify_set(current_edges),
        }
        if simple:
            cc["spanner"] = __intify_set(current_spanner)  # Simple
            free_edges = current_edges - current_spanner
            for edge in free_edges:
                pair = neighbors(edge, model.incidence_csc)
                if spin[pair[0]] == spin[pair[1]]:
                    cc["pivot"] = edge if edge is False else int(edge)
                    free_edges.discard(edge)
                    break
            else:
                cc["pivot"] = False
            cc["seeds"] = __intify_set(free_edges)
            injective = len(free_edges) == 0
            surjective = cc["pivot"] is not False
            cc["type"] = status_names_simple_connected[(injective, surjective)]
        res.append(cc)
    return res


class Model:
    """
    Main class to manipulate stochatic matching models.

    Parameters
    ----------
    incidence: :class:`~numpy.ndarray` or :class:`list`, optional
        Incidence matrix. If used, the graph will be considered as a hypergraph.
    adjacency: :class:`~numpy.ndarray` or :class:`list`, optional
        Adjacency matrix. If used, the graph will be considered as a simple graph.
    rates: :class:`~numpy.ndarray` or :class:`list` or :class:`str`, optional
        Arrival rates. You can use a specific rate vector or list.
        You can use `uniform` or `proportional` for uniform or degree-proportional allocation.
        Default to `proportional`, which makes the problem stabilizable if the graph is surjective.
    names: :class:`list` of :class:`str` or 'alpha', optional
        List of node names (e.g. for display)
    tol: :class:`float`, optional
        Values of absolute value lower than `tol` are set to 0.

    Examples
    --------

    The following examples are about stability:

    Is a triangle that checks triangular inequality stable?

    >>> import stochastic_matching.graphs as sm
    >>> triangle = sm.Cycle(rates="uniform")
    >>> triangle.stabilizable
    True

    >>> triangle.kernel.type
    'Bijective'

    We can look at the base flow (based on Moore-Penrose inverse by default).

    >>> triangle.base_flow
    array([0.5, 0.5, 0.5])

    As the graph is bijective, all optimizations will yield the same flow.

    >>> triangle.incompressible_flow()
    array([0.5, 0.5, 0.5])

    >>> triangle.optimize_edge(0)
    array([0.5, 0.5, 0.5])

    What if the triangular inequality does not hold?

    >>> triangle.rates = [1, 3, 1]
    >>> triangle.stabilizable
    False

    We can look at the base flow (based on Moore-Penrose inverse).

    >>> triangle.base_flow
    array([ 1.5, -0.5,  1.5])

    Now a bipartite example.

    >>> banner = sm.Tadpole(m=4, rates='proportional')

    Notice that we have a perfectly working solution with respect to conservation law.

    >>> banner.base_flow
    array([1., 1., 1., 1., 1.])

    However, the left kernel is not trivial.

    >>> banner.kernel.left
    array([[ 1],
           [-1],
           [ 1],
           [-1],
           [ 1]])

    As a consequence, stability is False.

    >>> banner.stabilizable
    False

    >>> banner.kernel.type
    'Nonjective'

    Note that the base flow can be negative even if there is a positive solution.

    >>> diamond = sm.CycleChain(rates=[5, 6, 2, 1])
    >>> diamond.base_flow
    array([ 3.5,  1.5,  1. ,  1.5, -0.5])
    >>> diamond.stabilizable
    True
    >>> diamond.maximin.round(decimals=6)
    array([4.5, 0.5, 1. , 0.5, 0.5])

    >>> diamond.incompressible_flow()
    array([4., 0., 1., 0., 0.])

    >>> diamond.kernel.type
    'Surjective-only'
    """

    name = "Generic"

    def __init__(
        self, incidence=None, adjacency=None, rates=None, names=None, tol=1e-7
    ):
        self.tol = tol
        self.base_flow = None
        self.__seeds = None
        self.incidence = incidence
        self.adjacency = adjacency
        self.rates = rates
        self.names = names
        self.simulator = None
        self.simulation = None

    @property
    def names(self):
        """
        :class:`list` of :class:`str`: list of node names (e.g. for display).

        If set to "alpha", automatic alphabetic labeling is used.
        If set to None, numeric labeling is used.
        """
        return self.__names

    @names.setter
    def names(self, names):
        if type(names) == str and names == "alpha":
            self.__names = CharMaker()
        else:
            self.__names = names

    @property
    def incidence(self):
        """
        :class:`~numpy.ndarray`: Incidence matrix of the graph.

        Setting incidence treats the graph as hypergraph.
        """
        return self._incidence.toarray().astype(int)

    @property
    def incidence_csr(self):
        """
        :class:`~scipy.sparse.csr_matrix`: Incidence matrix of the graph in csr format.
        """
        return self._incidence

    @property
    def incidence_csc(self):
        """
        :class:`~scipy.sparse.csc_matrix`: Incidence matrix of the graph in csc format.
        """
        return self._co_incidence

    @incidence.setter
    def incidence(self, incidence):
        if incidence is None:
            self._incidence = None
        else:
            if type(incidence) == list:
                incidence = np.array(incidence).astype(int)
            self._incidence = sp.csr_matrix(incidence)
            self._co_incidence = sp.csc_matrix(incidence)
            self._adjacency = None

    @property
    def adjacency(self):
        """
        :class:`~numpy.ndarray`: Adjacency matrix of the graph.

        Setting adjacency treats the graph as simple.
        """
        return self._adjacency

    @adjacency.setter
    def adjacency(self, adjacency):
        if adjacency is not None:
            adjacency = np.array(adjacency)
            incidence = adjacency_to_incidence(adjacency)
            self._incidence = sp.csr_matrix(incidence)
            self._co_incidence = sp.csc_matrix(incidence)
        self._adjacency = adjacency

    @property
    def n(self):
        """
        :class:`int`: Number of nodes.
        """
        return self._incidence.shape[0]

    @property
    def m(self):
        """
        :class:`int`: Number of edges.
        """
        return self._incidence.shape[1]

    @cached_property
    def degree(self):
        """
        :class:`~numpy.ndarray`: Degree vector.
        """
        return self._incidence @ np.ones(self.m)

    @cached_property
    def kernel(self):
        """
        :class:`~stochastic_matching.model.Kernel`: Description of the kernels of the incidence.
        """
        kernel = Kernel(self.incidence)
        if self.adjacency is not None:
            self.__seeds = [
                int(i) for c in self.connected_components for i in c["seeds"]
            ]
            kernel.right = simple_right_kernel(kernel.right, self.__seeds)
            kernel.right_inverse = kernel_inverse(kernel.right)
            kernel.left = simple_left_kernel(kernel.left)
        return kernel

    @cached_property
    def connected_components(self):
        """
        :class:`list` of :class:`dict`: description of the connected components of the graph.
        """
        return traversal(self)

    @property
    def seeds(self):
        """
        :class:`list` of :class:`int`: edges that induce a description of the (right) kernel as cycles and paddles.

        Changing the seeds modifies the right kernel accordingly. Only available on simple graphs.
        """
        if self.adjacency is not None:
            return self.__seeds

    @seeds.setter
    def seeds(self, seeds):
        if self.adjacency is not None:
            self.kernel.right = simple_right_kernel(self.kernel.right, seeds)
            self.kernel.right_inverse = kernel_inverse(self.kernel.right)
            self.__seeds = seeds

    @property
    def rates(self):
        """
        :class:`~numpy.ndarray`: vector of arrival rates.

        You can use `uniform` or `proportional` for uniform or degree-proportional allocation.
        Default to `proportional`, which makes the problem stabilizable if the graph is bijective.
        """
        return self.__rates

    @rates.setter
    def rates(self, mu):
        if mu is None:
            mu = self.degree
        if mu is not None:
            if isinstance(mu, str):
                if mu == "uniform":
                    mu = np.ones(self.n)
                else:
                    mu = self.degree
            else:
                mu = np.array(mu)
        self.__rates = mu
        self.__moore_penrose = None
        self.__maximin = None
        self.__vertices = None
        self.base_flow = self.moore_penrose

    @property
    def moore_penrose(self):
        """
        :class:`~numpy.ndarray`: Solution of the conservation equation obtained using the Moore-Penrose inverse.
        """
        if self.__moore_penrose is None:
            self.__moore_penrose = self.kernel.inverse @ self.rates
            clean_zeros(self.__moore_penrose, tol=self.tol)
        return self.__moore_penrose

    def edge_to_kernel(self, edge):
        """
        Parameters
        ----------
        edge: :class:`~numpy.ndarray`
            A flow vector in edge coordinates.

        Returns
        -------
        :class:`~numpy.ndarray`
            The same flow vector in kernel coordinates, based on the current base flow and right kernel.

        Examples
        --------

        Consider the codomino graph with a kernel with a kayak paddle.

        >>> import stochastic_matching.graphs as sm
        >>> codomino = sm.Codomino(rates = [3, 12, 3, 3, 12, 3])

        Default seeds of the codomino:

        >>> codomino.seeds
        [5, 7]

        Let use other seeds.

        >>> codomino.seeds = [0, 4]
        >>> codomino.kernel.right  # doctest: +NORMALIZE_WHITESPACE
        array([[ 1, -1,  1, -2,  0,  1, -1,  1],
           [ 0,  0, -1,  1,  1, -1,  0,  0]])

        Consider the Moore-Penrose flow and the maximin flow.

        >>> codomino.moore_penrose
        array([3., 0., 3., 6., 0., 3., 0., 3.])

        >>> codomino.maximin
        array([2., 1., 1., 9., 1., 1., 1., 2.])

        As the Moore-Penrose is the default base flow, its coordinates are obviously null.

        >>> codomino.edge_to_kernel(codomino.moore_penrose)
        array([0., 0.])

        As for maximin, one can check that the following kernel coordinates transform Moore-Penrose into it:

        >>> codomino.edge_to_kernel(codomino.maximin)
        array([-1.,  1.])

        If we change the base flow to maximin, we will see the coordinates shifted by (1, -1):

        >>> codomino.base_flow = codomino.maximin
        >>> codomino.edge_to_kernel(codomino.moore_penrose)
        array([ 1., -1.])
        >>> codomino.edge_to_kernel(codomino.maximin)
        array([0., 0.])
        """
        res = (self.kernel.right @ (edge - self.base_flow)) @ self.kernel.right_inverse
        clean_zeros(res, tol=self.tol)
        return res

    def kernel_to_edge(self, alpha):
        """
        Parameters
        ----------
        alpha: :class:`~numpy.ndarray` ot :class:`list`
            A flow vector in kernel coordinates.

        Returns
        -------
        :class:`~numpy.ndarray`
            The same flow vector in edge coordinates, based on the current base flow and right kernel.

        Examples
        --------

        Consider the codomino graph with a kernel with a kayak paddle.

        >>> import stochastic_matching.graphs as sm
        >>> codomino = sm.Codomino(rates=[3, 12, 3, 3, 12, 3])
        >>> codomino.kernel.right # doctest: +NORMALIZE_WHITESPACE
        array([[ 0,  0,  1, -1, -1,  1,  0,  0],
           [ 1, -1,  0, -1,  1,  0, -1,  1]])

        Consider the Moore-Penrose and the maximin flows.

        >>> codomino.moore_penrose
        array([3., 0., 3., 6., 0., 3., 0., 3.])

        >>> codomino.maximin
        array([2., 1., 1., 9., 1., 1., 1., 2.])

        As the Moore-Penrose inverse is the base flow, it is (0, 0) in kernel coordinates.

        >>> codomino.kernel_to_edge([0, 0])
        array([3., 0., 3., 6., 0., 3., 0., 3.])

        As for maximin, (-2, -1) seems to be its kernel coordinates.

        >>> codomino.kernel_to_edge([-2, -1])
        array([2., 1., 1., 9., 1., 1., 1., 2.])
        >>> np.allclose(codomino.kernel_to_edge([-2, -1]), codomino.maximin)
        True

        If we change the kernel space, the kernel coordinates change as well.

        >>> codomino.seeds = [0, 4]
        >>> codomino.kernel.right # doctest: +NORMALIZE_WHITESPACE
        array([[ 1, -1,  1, -2,  0,  1, -1,  1],
           [ 0,  0, -1,  1,  1, -1,  0,  0]])
        >>> codomino.kernel_to_edge([0, 0])
        array([3., 0., 3., 6., 0., 3., 0., 3.])

        >>> codomino.kernel_to_edge([-2, -1])
        array([ 1.,  2.,  2.,  9., -1.,  2.,  2.,  1.])
        >>> np.allclose(codomino.kernel_to_edge([-2, -1]), codomino.maximin)
        False
        """
        res = (alpha @ self.kernel.right) + self.base_flow
        clean_zeros(res, tol=self.tol)
        return res

    @property
    def maximin(self):
        """
        :class:`~numpy.ndarray`: solution of the conservation equation that maximizes the minimal flow over all edges.
        """
        if self.__maximin is None:
            d, m = self.kernel.right.shape
            if d == 0:
                self.__maximin = self.moore_penrose
            else:
                # Better in theory but better precision with the legacy approach
                # c = np.zeros(m + 1)
                # c[-1] = -1
                # a_ub = sp.hstack([-sp.identity(m), sp.csr_matrix(np.ones((m, 1)))])
                # b_ub = np.zeros(m)
                # a_eq = sp.csr_matrix((self.incidence_csr.data,
                #                    self.incidence_csr.indices,
                #                    self.incidence_csr.indptr), shape=(self.n, m + 1))
                # b_eq = self.rates
                # optimizer = linprog(c=c,
                #                     A_ub=a_ub,
                #                     b_ub=b_ub,
                #                     A_eq=a_eq,
                #                     b_eq=b_eq,
                #                     bounds=[(None, None)] * (m + 1),
                #                     options={'sparse': True}
                #                     )
                # flow = optimizer.x[:-1]

                c = np.zeros(d + 1)
                c[d] = 1
                a_ub = -np.vstack([self.kernel.right, np.ones(m)]).T
                optimizer = linprog(
                    c=c,
                    A_ub=a_ub,
                    b_ub=self.moore_penrose,
                    bounds=[(None, None)] * (d + 1),
                )
                flow = optimizer.slack - optimizer.x[-1]
                clean_zeros(flow, tol=self.tol)
                self.__maximin = flow
        return self.__maximin

    def shadow_prices(self, rewards):
        """
        Return the shadow prices of each item type for a given reward. Cf https://doi.org/10.1145/3578338.3593532

        Parameters
        ----------
        rewards: :class:`~numpy.ndarray` or :class:`list`
            Rewards associated to each edge.

        Returns
        -------
        :class:`~numpy.ndarray`
            Shadow prices.
        """
        optimizer = linprog(
            c=self.rates,
            A_ub=-self.incidence.T,
            b_ub=-np.array(rewards),
            bounds=[(None, None)] * self.n
        )
        clean_zeros(optimizer.x, tol=self.tol)
        return optimizer.x

    def normalize_rewards(self, rewards):
        """
        Parameters
        ----------
        rewards: :class:`~numpy.ndarray` or :class:`list`
            Rewards associated to each edge.

        Returns
        -------
        :class:`~numpy.ndarray`
            For bijective cases, rewards with negative values on taboo edges, null values everywhere else.

        Examples
        --------

        >>> import stochastic_matching as sm
        >>> diamond = sm.CycleChain(rates=[4, 5, 2, 1])

        First edge provides no reward (bijective vertex).

        >>> diamond.normalize_rewards([0, 1, 1, 1, 1])
        array([ 0.,  0.,  0.,  0., -1.])

        First edge provides more reward (injective-only vertex).

        >>> diamond.normalize_rewards([2, 1, 1, 1, 1])
        array([ 0.,  0.,  0., -1.,  0.])

        .. note::
            Reward normalization can miss a taboo edge in injective-only case.

        On bijective graphs, all edges are OK.

        >>> paw = sm.Tadpole()
        >>> paw.normalize_rewards([6, 3, 1, 2])
        array([0., 0., 0., 0.])

        Last example: the hypergraph from Nazari & Stolyar.

        >>> ns = sm.NS19(rates=[1.2, 1.5, 2, 0.8])
        >>> ns.normalize_rewards([-1, -1, 1, 2, 5, 4, 7])
        array([-2., -5.,  0.,  0.,  0., -1.,  0.])
        """
        return rewards - self.shadow_prices(rewards) @ self.incidence

    def gentle_rewards(self, rewards):
        """
        Parameters
        ----------
        rewards: :class:`~numpy.ndarray` or :class:`list`
            Rewards associated to each edge.

        Returns
        -------
        :class:`~numpy.ndarray`
            Rewards with negative values on taboo edges, positive values everywhere else.
            Absolute values are degree proportional.

        Examples
        --------

        >>> import stochastic_matching as sm
        >>> diamond = sm.CycleChain(rates=[4, 5, 2, 1])

        First edge provides no reward (bijective vertex).

        >>> diamond.gentle_rewards([0, 1, 1, 1, 1])
        array([ 2,  2,  2,  2, -2])

        First edge provides more reward (injective-only vertex).

        >>> diamond.gentle_rewards([2, 1, 1, 1, 1])
        array([ 2, -2,  2, -2,  2])

        On bijective graphs, all edges are OK.

        >>> paw = sm.Tadpole()
        >>> paw.gentle_rewards([6, 3, 1, 2])
        array([2, 2, 2, 2])

        Last example: the hypergraph from Nazari & Stolyar.

        >>> ns = sm.NS19(rates=[1.2, 1.5, 2, 0.8])
        >>> ns.gentle_rewards([-1, -1, 1, 2, 5, 4, 7])
        array([-1, -1,  1,  1,  2, -2,  3])
        """
        alt_rewards = self.incidence.sum(axis=0)
        mask = self.optimize_rates(rewards) == 0
        alt_rewards[mask] = -alt_rewards[mask]
        return alt_rewards

    def optimize_rates(self, weights):
        """
        Tries to find a positive solution that minimizes/maximizes a given edge.

        Parameters
        ----------
        weights: :class:`~numpy.ndarray` or :class:`list`
            Rewards associated to each edge.

        Returns
        -------
        :class:`~numpy.ndarray`
            Optimized flow that maximize the total reward.

        Examples
        --------

        >>> import stochastic_matching as sm
        >>> diamond = sm.CycleChain(rates=[4, 5, 2, 1])

        Optimize with the first edge that provides no reward.

        >>> diamond.optimize_rates([0, 1, 1, 1, 1])
        array([3., 1., 1., 1., 0.])

        Optimize with the first edge that provides more reward.

        >>> diamond.optimize_rates([2, 1, 1, 1, 1])
        array([4., 0., 1., 0., 1.])

        On bijective graphs, the method directly returns the unique solution.

        >>> paw = sm.Tadpole()
        >>> paw.optimize_rates([6, 3, 1, 2])
        array([1., 1., 1., 1.])

        Last example: the hypergraph from Nazari & Stolyar.

        >>> ns = sm.NS19(rates=[1.2, 1.5, 2, 0.8])
        >>> ns.optimize_rates([-1, -1, 1, 2, 5, 4, 7])
        array([0. , 0. , 1.7, 0.5, 1.2, 0. , 0.3])
        """
        d, m = self.kernel.right.shape
        if d == 0:
            return self.moore_penrose
        else:
            weights = np.array(weights)
            c = -self.kernel.right @ weights
            optimizer = linprog(
                c=c,
                A_ub=-self.kernel.right.T,
                b_ub=self.moore_penrose,
                bounds=[(None, None)] * d,
            )
            clean_zeros(optimizer.slack, tol=self.tol)
            return optimizer.slack

    def optimize_edge(self, edge, sign=1):
        """
        Tries to find a positive solution that minimizes/maximizes a given edge.

        Parameters
        ----------
        edge: :class:`int`
            Edge to optimize.
        sign: :class:`int`
            Use 1 to maximize, -1 to minimize.

        Returns
        -------
        :class:`~numpy.ndarray`
            Optimized flow.
        """
        d, m = self.kernel.right.shape
        if d == 0:
            return self.moore_penrose
        else:
            weights = np.zeros(self.m)
            weights[edge] = sign
            return self.optimize_rates(weights)

    def incompressible_flow(self):
        """
        Finds the minimal flow that must pass through each edge. This is currently done in a *brute force* way
        by minimizing every edges.

        Returns
        -------
        :class:`~numpy.ndarray`
            Unavoidable flow.
        """
        d, m = self.kernel.right.shape
        if d == 0:
            return self.base_flow
        else:
            flow = np.zeros(m)
            for edge in range(m):
                flow[edge] = self.optimize_edge(edge, -1)[edge]
            clean_zeros(flow, tol=self.tol)
            return flow

    @property
    def stabilizable(self):
        """
        :class:`bool`: Is the model stabilizable, i.e. is it bijective with a positive solution of the conservation law.
        """
        return bool(np.all(self.maximin > 0) and self.kernel.status[1])

    @property
    def vertices(self):
        """
        :class:`list`: list of vertices

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> paw = sm.Tadpole()
        >>> paw.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': None, 'edge_coordinates': array([1., 1., 1., 1.]),
          'null_edges': [], 'bijective': True}]

        Note that vertices is soft-cached: as long as the graph and rates do not change they are not re-computed.

        >>> paw.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': None, 'edge_coordinates': array([1., 1., 1., 1.]),
          'null_edges': [], 'bijective': True}]


        >>> star = sm.Star()
        >>> star.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': None, 'edge_coordinates': array([1., 1., 1.]),
          'null_edges': [], 'bijective': False}]

        >>> cycle = sm.Cycle(4, rates=[3, 3, 3, 1])
        >>> cycle.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([-0.75]),
          'edge_coordinates': array([1. , 1.5, 2.5, 0. ]),
          'null_edges': [3], 'bijective': False},
         {'kernel_coordinates': array([0.75]), 'edge_coordinates': array([2.5, 0. , 1. , 1.5]),
          'null_edges': [1], 'bijective': False}]

        >>> diamond = sm.CycleChain(rates=[3, 3, 3, 1])
        >>> diamond.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([-0.5]), 'edge_coordinates': array([1., 2., 1., 1., 0.]),
          'null_edges': [4], 'bijective': True},
         {'kernel_coordinates': array([0.5]), 'edge_coordinates': array([2., 1., 1., 0., 1.]),
         'null_edges': [3], 'bijective': True}]

        >>> diamond.rates = [2, 3, 2, 1]
        >>> diamond.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([-0.25]), 'edge_coordinates': array([1., 1., 1., 1., 0.]),
          'null_edges': [4], 'bijective': True},
         {'kernel_coordinates': array([0.75]), 'edge_coordinates': array([2., 0., 1., 0., 1.]),
          'null_edges': [1, 3], 'bijective': False}]

        >>> diamond.rates = [1, 3, 1, 1]
        >>> diamond.vertices  # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        ValueError: The matching model admits no positive solution.

        >>> diamond.rates = None
        >>> diamond.vertices  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([-1.]), 'edge_coordinates': array([0., 2., 1., 2., 0.]),
          'null_edges': [0, 4], 'bijective': False},
         {'kernel_coordinates': array([1.]), 'edge_coordinates': array([2., 0., 1., 0., 2.]),
          'null_edges': [1, 3], 'bijective': False}]

        >>> codomino = sm.Codomino()
        >>> codomino.rates = [4, 5, 5, 3, 3, 2]
        >>> codomino.seeds = [1, 2]
        >>> codomino.base_flow = codomino.maximin
        >>> sorted(codomino.vertices, key=str)  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([ 1., -1.]), 'edge_coordinates': array([1., 3., 1., 3., 1., 0., 2., 0.]),
        'null_edges': [5, 7], 'bijective': True},
        {'kernel_coordinates': array([-1.,  0.]), 'edge_coordinates': array([3., 1., 2., 0., 2., 1., 0., 2.]),
        'null_edges': [3, 6], 'bijective': True},
        {'kernel_coordinates': array([-1., -1.]), 'edge_coordinates': array([3., 1., 1., 1., 3., 0., 0., 2.]),
        'null_edges': [5, 6], 'bijective': True},
        {'kernel_coordinates': array([0., 1.]), 'edge_coordinates': array([2., 2., 3., 0., 0., 2., 1., 1.]),
        'null_edges': [3, 4], 'bijective': True},
        {'kernel_coordinates': array([1., 0.]), 'edge_coordinates': array([1., 3., 2., 2., 0., 1., 2., 0.]),
        'null_edges': [4, 7], 'bijective': True}]

        >>> codomino.rates = [2, 4, 2, 2, 4, 2]
        >>> codomino.base_flow = np.array([1.0, 1.0, 1.0, 2.0, 0.0, 1.0, 1.0, 1.0])
        >>> sorted(codomino.vertices, key=str)  # doctest: +NORMALIZE_WHITESPACE
        [{'kernel_coordinates': array([ 1., -1.]), 'edge_coordinates': array([0., 2., 0., 4., 0., 0., 2., 0.]),
        'null_edges': [0, 2, 4, 5, 7], 'bijective': False},
        {'kernel_coordinates': array([-1.,  1.]), 'edge_coordinates': array([2., 0., 2., 0., 0., 2., 0., 2.]),
        'null_edges': [1, 3, 4, 6], 'bijective': False},
        {'kernel_coordinates': array([-1., -1.]), 'edge_coordinates': array([2., 0., 0., 2., 2., 0., 0., 2.]),
        'null_edges': [1, 2, 5, 6], 'bijective': False}]

        >>> pyramid = sm.Pyramid(rates=[4, 3, 3, 3, 6, 6, 3, 4, 4, 4])
        >>> pyramid.seeds = [0, 12, 2]
        >>> pyramid.base_flow = pyramid.kernel_to_edge([1/6, 1/6, 1/6] )
        >>> sorted([(v['kernel_coordinates'], v['bijective']) for v in pyramid.vertices], key=str) # doctest: +NORMALIZE_WHITESPACE
        [(array([ 1., -1.,  0.]), True),
         (array([-1.,  1.,  0.]), True),
         (array([-1., -1.,  0.]), True),
         (array([0., 0., 1.]), False),
         (array([1., 1., 0.]), True)]
        """
        if self.__vertices is not None:
            return self.__vertices
        if not np.all(self.maximin > 0):
            raise ValueError("The matching model admits no positive solution.")
        d, m = self.kernel.right.shape
        if d == 0:
            dico = {
                "kernel_coordinates": None,
                "edge_coordinates": self.maximin,
                "null_edges": [i for i in range(m) if self.maximin[i] == 0],
            }
            dico["bijective"] = (self.m - self.n) == len(dico["null_edges"])
            res = [dico]
        else:
            if d > 1:
                halfspaces = np.hstack(
                    [-self.kernel.right.T, -self.base_flow.reshape(self.m, 1)]
                )
                interior_point = self.edge_to_kernel(self.maximin)
                hs = HalfspaceIntersection(halfspaces, interior_point)
                verts = hs.intersections
                clean_zeros(verts)
                verts = np.unique(verts, axis=0)
            else:
                verts = [None, None]
                for e in range(m):
                    spin = self.kernel.right[0, e]
                    if spin == 0:
                        continue
                    v = -self.base_flow[e] / spin
                    if spin > 0:
                        if verts[0] is None:
                            verts[0] = v
                        else:
                            verts[0] = max(v, verts[0])
                    else:
                        if verts[1] is None:
                            verts[1] = v
                        else:
                            verts[1] = min(v, verts[1])
                verts = np.array(verts).reshape((2, 1))
            v = verts.shape[0]
            res = [dict() for _ in range(v)]
            for i, dico in enumerate(res):
                dico["kernel_coordinates"] = verts[i, :]
                dico["edge_coordinates"] = self.kernel_to_edge(
                    dico["kernel_coordinates"]
                )
                clean_zeros(dico["edge_coordinates"], tol=self.tol)
                dico["null_edges"] = [
                    i for i in range(m) if dico["edge_coordinates"][i] == 0
                ]
                dico["bijective"] = (self.m - self.n) == len(dico["null_edges"])
        self.__vertices = res
        return res

    def show(self, **kwargs):
        """
        Shows the model. See :meth:`~stochastic_matching.display.show` for details.

        Parameters
        ----------
        **kwargs
            See :meth:`~stochastic_matching.display.show` for details.

        Returns
        -------
        :class:`~IPython.display.HTML`

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> pyramid = sm.Pyramid()
        >>> pyramid.show()
        <IPython.core.display.HTML object>
        """
        show(self, **kwargs)

    def show_graph(self, **kwargs):
        """
        Shows the graph of the model (with node names and no value on edges by default).

        Parameters
        ----------
        **kwargs
            See :meth:`~stochastic_matching.display.show` for details.

        Returns
        -------
        :class:`~IPython.display.HTML`

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> pyramid = sm.Pyramid()
        >>> pyramid.show_graph()
        <IPython.core.display.HTML object>
        """
        default = {"disp_flow": False, "disp_rates": False}
        show(self, **{**default, **kwargs})

    def show_flow(self, **kwargs):
        """
        Shows the model (with check on conservation law and edge positivity).

        Parameters
        ----------
        **kwargs
            See :meth:`~stochastic_matching.display.show` for details.

        Returns
        -------
        :class:`~IPython.display.HTML`

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> pyramid = sm.Pyramid()
        >>> pyramid.show_flow()
        <IPython.core.display.HTML object>
        """
        default = {"check_flow": True}
        show(self, **{**default, **kwargs})

    def show_kernel(self, **kwargs):
        """
        Shows the kernel basis.

        Parameters
        ----------
        **kwargs
            See :meth:`~stochastic_matching.display.show` for details.

        Returns
        -------
        :class:`~IPython.display.HTML`

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> pyramid = sm.Pyramid()
        >>> pyramid.show_kernel()
        <IPython.core.display.HTML object>
        """
        default = {"disp_kernel": True, "disp_flow": False, "disp_zero": False}
        show(self, **{**default, **kwargs})

    def show_vertex(self, i, **kwargs):
        """
        Shows the vertex of indice *i*. See :meth:`~stochastic_matching.display.show` for details.

        Parameters
        ----------
        i: :class:`int`
            indice of the vertex.
        **kwargs
            See :meth:`~stochastic_matching.display.show` for details.

        Returns
        -------
        :class:`~IPython.display.HTML`

        Examples
        --------

        >>> import stochastic_matching.graphs as sm
        >>> pyramid = sm.Pyramid()
        >>> pyramid.show_vertex(2)
        <IPython.core.display.HTML object>
        """
        flow = self.vertices[i]["edge_coordinates"]
        default = {"flow": flow, "disp_zero": False, "check_flow": True}
        show(self, **{**default, **kwargs})

    def run(self, simulator, n_steps=1000000, seed=None, max_queue=1000, **kwargs):
        """
        All-in-one instantiate and run simulation.

        Parameters
        ----------
        simulator: :class:`str` or :class:`~stochastic_matching.simulator.simulator.Simulator`
            Type of simulator to instantiate.
        n_steps: :class:`int`, optional
            Number of arrivals to simulate.
        seed: :class:`int`, optional
            Seed of the random generator
        max_queue: :class:`int`
            Max queue size. Necessary for speed and detection of unstability.
            For stable systems very close to the unstability
            border, the max_queue may be reached.
        **kwargs
            Arguments to pass to the simulator.

        Returns
        -------
        bool
            Success of simulation.

        Examples
        --------

        Let start with a working triangle and a greedy simulator.

        >>> import stochastic_matching.graphs as sm
        >>> triangle = sm.Cycle(rates=[3, 4, 5])
        >>> triangle.base_flow
        array([1., 2., 3.])
        >>> triangle.run('random_edge', seed=42, n_steps=20000)
        True
        >>> triangle.simulation
        array([1.0716, 1.9524, 2.976 ])

        A ill diamond graph (simulation ends before completion due to drift).

        Note that the drift is slow, so if the number of steps is low the simulation may complete without overflowing.

        >>> diamond = sm.CycleChain(rates='uniform')
        >>> diamond.base_flow
        array([0.5, 0.5, 0. , 0.5, 0.5])

        >>> diamond.run('longest', seed=42, n_steps=20000)
        True
        >>> diamond.simulation
        array([0.501 , 0.4914, 0.0018, 0.478 , 0.5014])

        A working candy. While candies are not good for greedy policies, the virtual queue is
        designed to deal with it.

        >>> candy = sm.HyperPaddle(rates=[1, 1, 1.1, 1, 1.1, 1, 1])
        >>> candy.base_flow
        array([0.95, 0.05, 0.05, 0.05, 0.05, 0.95, 1.  ])

        The above states that the target flow for the hyperedge of the candy (last entry) is 1.

        >>> candy.run('longest', seed=42, n_steps=20000)
        False
        >>> candy.simulator.logs.steps_done
        10458
        >>> candy.simulation  # doctest: +NORMALIZE_WHITESPACE
        array([0.64234079, 0.37590361, 0.38760757, 0.40757315, 0.40895009,
           0.59208262, 0.2939759 ])

        A greedy simulator performs poorly on the hyperedge.

        >>> candy.run('virtual_queue', seed=42, n_steps=20000)
        True
        >>> candy.simulation  # doctest: +NORMALIZE_WHITESPACE
        array([0.96048, 0.04104, 0.04428, 0.06084, 0.06084, 0.94464, 0.9846 ])

        The virtual queue simulator manages to cope with the target flow on the hyperedge.
        """
        simulator = class_converter(simulator, Simulator)
        self.simulator = simulator(
            self, n_steps=n_steps, seed=seed, max_queue=max_queue, **kwargs
        )
        self.simulator.run()
        self.simulation = self.simulator.flow
        self.base_flow = self.simulation
        return bool(
            self.simulator.internal["n_steps"] == self.simulator.logs.steps_done
        )



import numpy as np

from stochastic_matching.simulator.simulator import Simulator


class ExtendedSimulator(Simulator):
    """
    This extended (abstract) class of :class:`~stochastic_matching.simulator.simulator.Simulator` accepts
    additional parameters designed to reach vertices of the stability polytope.

    How these parameters are exactly used is determined by its sub-classes.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    rewards: :class:`~numpy.ndarray` or :class:`list`, optional
        Rewards associated to edges.
    alt_rewards: :class:`str`, optional
        Possible reward transformation.
    beta: :class:`float`, optional
        Reward factor: each edge is given a default score 1/beta (small beta means higher reward impact).
        Setting beta triggers reward-based scoring.
    forbidden_edges: :class:`list` or :class:`~numpy.ndarray`, or :class:`bool`, optional
        Edges that should not be used. If set to True, the list of forbidden edges is computed from the rewards.
    k: :class:`int`, optional
        Limit on queue size to apply edge interdiction (enforce stability on injective-only vertices).
    kwargs: :class:`dict`
        Keyword parameters of :class:`~stochastic_matching.simulator.simulator.Simulator`
    """

    def __init__(
        self, model, rewards=None, alt_rewards=None, beta=None, forbidden_edges=None, k=None, **kwargs
    ):
        if rewards is not None:
            self.rewards = np.array(rewards)
        else:
            self.rewards = model.incidence.sum(axis=0)
        if alt_rewards is None:
            self.alt_rewards = self.rewards
        elif alt_rewards == "gentle":
            self.alt_rewards = model.gentle_rewards(self.rewards)
        elif alt_rewards == "normalize":
            self.alt_rewards = model.normalize_rewards(self.rewards)
        else:
            raise ValueError(f"Unknown reward transformation: {alt_rewards}.")
        self.beta = beta
        if forbidden_edges is True:
            flow = model.optimize_rates(self.rewards)
            forbidden_edges = [i for i in range(model.m) if flow[i] == 0]
            if len(forbidden_edges) == 0:
                forbidden_edges = None
        self.forbidden_edges = forbidden_edges
        if forbidden_edges and rewards is None:
            self.rewards[forbidden_edges] = - self.rewards[forbidden_edges]
        self.k = k
        super().__init__(model, **kwargs)

    def set_internal(self):
        super().set_internal()
        if self.beta is None:
            scores = np.zeros(self.model.m, dtype=int)
        else:
            scores = self.alt_rewards / self.beta
        self.internal["scores"] = scores
        self.internal["forbidden_edges"] = self.forbidden_edges
        self.internal["k"] = self.k



from numba import njit
import numpy as np

from stochastic_matching.simulator.extended import ExtendedSimulator


@njit
def longest_core(
    logs,
    arrivals,
    graph,
    n_steps,
    queue_size,  # Generic arguments
    scores,
    forbidden_edges,
    k,
):
    """
    Jitted function for policies based on longest-queue first.

    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables.
    arrivals: :class:`~stochastic_matching.simulator.arrivals.Arrivals`
        Item arrival process.
    graph: :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Model graph.
    n_steps: :class:`int`
        Number of arrivals to process.
    queue_size: :class:`~numpy.ndarray`
        Number of waiting items of each class.
    scores: :class:`~numpy.ndarray`
        Default value for edges. Enables EGPD-like selection mechanism.
    forbidden_edges: :class:`list`, optional
        Edges that are disabled.
    k: :class:`int`, optional
        Queue size above which forbidden edges become available again.

    Returns
    -------
    None
    """

    n, max_queue = logs.queue_log.shape
    if k is None:
        k = max_queue + 1

    # Optimize forbidden edges and set greedy flag.
    greedy = False
    if forbidden_edges is not None:
        forbid = {k: True for k in forbidden_edges}
    else:
        forbid = {k: True for k in range(0)}
        if np.min(scores) > -1:
            greedy = True

    for age in range(n_steps):
        # Draw an arrival
        node = arrivals.draw()
        queue_size[node] += 1
        if queue_size[node] == max_queue:
            return None

        # Test if an actionable edge may be present
        best_edge = -1
        if not greedy or queue_size[node] == 1:
            # Should we activate edge restriction?
            if greedy:
                restrain = False
            else:
                restrain = True
                if k is not None:
                    for v in range(n):
                        if queue_size[v] >= k:
                            restrain = False
                            break

            best_score = 0
            # update scores
            for e in graph.edges(node):
                if restrain and e in forbid:
                    continue
                score = scores[e]
                for v in graph.nodes(e):
                    w = queue_size[v]
                    if w == 0:
                        break
                    if v != node:
                        score += w
                else:
                    if score > best_score:
                        best_edge = e
                        best_score = score

            if best_edge > -1:
                queue_size[graph.nodes(best_edge)] -= 1

        logs.update(queue_size=queue_size, node=node, edge=best_edge)


class Longest(ExtendedSimulator):
    """
    Matching simulator derived from :class:`~stochastic_matching.simulator.extended.ExtendedSimulator`.
    By default, the policy is greedy and whenever multiple edges are actionable,
    the longest queue (sum of queues of adjacent nodes) is chosen.

    Multiple options are available to tweak this behavior.

    Parameters
    ----------

    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    shift_rewards: :class:`bool`, default=True
        Longest only selects edges with a positive score. By default, all actionable edges have a positive score,
        which makes the policy greedy. However, if negative rewards are added into the mix, this is not True anymore.
        Setting shift_rewards to True ensures that default edge scores are non-negative so reward-based selection
        is greedy. Setting it to False enables a non-greedy reward-based selection.
    **kwargs
        Keyword parameters of :class:`~stochastic_matching.simulator.extended.ExtendedSimulator`.

    Examples
    --------

    Let's start with a working triangle. Not that the results are the same for all greedy simulator because
    there are no decision in a triangle (always at most one non-empty queue under a greedy policy).

    >>> import stochastic_matching as sm
    >>> sim = Longest(sm.Cycle(rates=[3, 4, 5]), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [287 338 375]
    Traffic: [125 162 213]
    Queues: [[838 104  41  13   3   1   0   0   0   0]
     [796 119  53  22   8   2   0   0   0   0]
     [640 176  92  51  24   9   5   3   0   0]]
    Steps done: 1000

    A non stabilizable diamond (simulation ends before completion due to drift).

    >>> sim = Longest(sm.CycleChain(rates='uniform'), n_steps=1000, seed=42, max_queue=10)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [85 82 85 86]
    Traffic: [38 38  7 37 40]
    Queues: [[126  74  28  37  21  32  16   1   2   1]
     [326   8   3   1   0   0   0   0   0   0]
     [321  12   4   1   0   0   0   0   0   0]
     [ 90  80  47  37  37  23  11   3   5   5]]
    Steps done: 338

    A stabilizable candy (but candies are not good for greedy policies).

    >>> sim = Longest(sm.HyperPaddle(rates=[1, 1, 1.5, 1, 1.5, 1, 1]), n_steps=1000, seed=42, max_queue=25)
    >>> sim.run()
    >>> sim.plogs # doctest: +NORMALIZE_WHITESPACE
    Arrivals: [46 26 32 37 70 35 45]
    Traffic: [24 17  2 23 33 12 13]
    Queues: [[ 23  32  45  38  22  43  31  34  20   3   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [290   1   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [290   1   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [ 9   1   7   9   3   3  26  37   4   8  10   9   2  10  40  11   2  16
        3   3  21  27  22   1   7]
     [212  49  22   5   3   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [233  41   6   7   4   0   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]
     [231  33  16   4   6   1   0   0   0   0   0   0   0   0   0   0   0   0
        0   0   0   0   0   0   0]]
    Steps done: 291

    Using greedy rewards-based longest (insprired by Stolyar EGCD technique), one can try to reach some vertices.

    That works for the Fish graph:

    >>> fish = sm.KayakPaddle(m=4, l=0, rates=[4, 4, 3, 2, 3, 2])
    >>> sim = Longest(fish, rewards=[0, 2, 2, 0, 1, 1, 0], beta=.01, n_steps=10000, seed=42)
    >>> sim.run()

    The flow avoids one edge:

    >>> sim.flow
    array([2.8764, 1.089 , 1.0008, 0.    , 0.8532, 2.079 , 1.0062])

    The price is one node having a big queue:

    >>> avg_queues = sim.avg_queues
    >>> avg_queues[-1]
    np.float64(61.1767)

    Other nodes are not affected:

    >>> np.round(np.mean(avg_queues[:-1]), decimals=4)
    np.float64(0.7362)

    Playing with the beta parameter allows to adjust the trade-off (smaller queue, leak on the forbidden edge):

    >>> sim = Longest(fish, rewards=[0, 2, 2, 0, 1, 1, 0], beta=.1, n_steps=10000, seed=42)
    >>> sim.run()
    >>> sim.flow
    array([2.9574, 1.008 , 0.9198, 0.018 , 0.9972, 2.061 , 1.0242])
    >>> sim.avg_queues[-1]
    np.float64(8.4628)

    Alternatively, one can use the k-filtering techniques:

    >>> diamond = sm.CycleChain(rates=[1, 2, 2, 1])
    >>> diamond.run('longest', forbidden_edges=[0, 4], seed=42,
    ...                            k=100, n_steps=1000, max_queue=1000)
    True
    >>> diamond.simulation
    array([0.   , 0.954, 0.966, 0.954, 0.   ])

    Same result can be achieved by putting low rewards on 0 and 4 and settings forbidden_edges to True.

    >>> diamond.run('longest', rewards=[1, 2, 2, 2, 1], seed=42, forbidden_edges=True,
    ...                            k=100, n_steps=1000, max_queue=1000)
    True
    >>> diamond.simulation
    array([0.   , 0.954, 0.966, 0.954, 0.   ])

    Note that if the threshold is too low some exceptions are done to limit the queue sizes.

    >>> diamond.run('longest', rewards=[1, 2, 2, 2, 1], seed=42, forbidden_edges=True,
    ...                            k=10, n_steps=1000, max_queue=1000)
    True
    >>> diamond.simulation
    array([0.108, 0.93 , 0.966, 0.93 , 0.024])

    Having no relaxation usually leads to large, possibly overflowing, queues.
    However, this doesn't show on small simulations.

    >>> diamond.run('longest', rewards=[1, 2, 2, 2, 1], seed=42, forbidden_edges=True,
    ...                            n_steps=1000, max_queue=100)
    True
    >>> diamond.simulator.avg_queues
    array([7.515, 7.819, 1.067, 1.363])
    >>> diamond.simulation
    array([0.   , 0.954, 0.966, 0.954, 0.   ])

    To compare with the priority-based pure greedy version:

    >>> diamond.run('priority', weights=[1, 2, 2, 2, 1], n_steps=1000, max_queue=1000, seed=42)
    True
    >>> diamond.simulation
    array([0.444, 0.63 , 0.966, 0.63 , 0.324])

    We get similar results with Stolyar greedy longest:

    >>> diamond.run('longest', rewards=[-1, 1, 1, 1, -1], beta=.01, n_steps=1000, max_queue=1000, seed=42)
    True
    >>> diamond.simulation
    array([0.444, 0.63 , 0.966, 0.63 , 0.324])

    However, if you remove the greedyness, you can converge to the vertex:

    >>> diamond.run('longest', rewards=[-1, 1, 1, 1, -1], beta=.01,
    ...             n_steps=1000, max_queue=1000, seed=42, shift_rewards=False)
    True
    >>> diamond.simulation
    array([0.   , 0.954, 0.966, 0.954, 0.   ])

    Another example with other rates.

    >>> diamond.rates=[4, 5, 2, 1]

    Optimize with the first and last edges that provide less reward.

    >>> diamond.run('longest', rewards=[1, 2, 2, 2, 1], seed=42, forbidden_edges=True,
    ...                            k=100, n_steps=1000, max_queue=1000)
    True
    >>> diamond.simulation
    array([3.264, 0.888, 0.948, 0.84 , 0.   ])

    Increase the reward on the first edge.

    >>> diamond.run('longest', rewards=[4, 2, 2, 2, 1], seed=42, forbidden_edges=True,
    ...                            k=100, n_steps=1000, max_queue=1000)
    True
    >>> diamond.simulation
    array([4.152, 0.   , 0.996, 0.   , 0.84 ])

    On bijective graphs, no edge is forbidden whatever the weights.

    >>> paw = sm.Tadpole()
    >>> paw.run('longest', rewards=[6, 3, 1, 2], seed=42, forbidden_edges=True,
    ...                            k=100, n_steps=1000, max_queue=1000)
    True
    >>> paw.simulation
    array([1.048, 1.056, 1.016, 0.88 ])
    """

    name = "longest"

    def __init__(self, model, shift_rewards=True, **kwargs):
        self.shift_rewards = shift_rewards
        super().__init__(model, **kwargs)

    def set_internal(self):
        super().set_internal()
        if self.shift_rewards:
            self.internal["scores"] -= np.min(self.internal["scores"])

    def run(self):
        longest_core(logs=self.logs, **self.internal)



from numba import int64
from numba.experimental import jitclass
import numpy as np


@jitclass
class Logs:
    """
    Jitclass for logs.

    Parameters
    ----------
    n: :class:`int`
        Number of nodes.
    m: :class:`int`
        Number of edges.
    max_queue: :class:`int`
        Maximum number of items per class.
    """

    income: int64[:]
    traffic: int64[:]
    queue_log: int64[:, :]

    def __init__(self, n, m, max_queue):
        self.income = np.zeros(n, dtype=np.int64)
        self.traffic = np.zeros(m, dtype=np.int64)
        self.queue_log = np.zeros((n, max_queue), dtype=np.int64)

    @property
    def steps_done(self):
        return int(np.sum(self.income))

    def update(self, queue_size, node, edge):
        self.income[node] += 1
        if edge > -1:
            self.traffic[edge] += 1
        for i, q in enumerate(queue_size):
            self.queue_log[i, q] += 1


@jitclass
class PhantomLogs:
    """
    Jitclass for logs of e-filtering policies. automatically concatenates the results in the original format.

    Parameters
    ----------
    n: :class:`int`
        Number of nodes of the original graph.
    m: :class:`int`
        Number of edges of the original graph.
    max_queue: :class:`int`
        Maximum number of items per class.
    edges: :class:`~numpy.ndarray`
        Edge index.
    """

    income: int64[:]
    traffic: int64[:]
    edges: int64[:]
    queue_log: int64[:, :]
    steps_done: int
    n: int

    def __init__(self, n, m, max_queue, edges):
        self.income = np.zeros(n, dtype=np.int64)
        self.traffic = np.zeros(m, dtype=np.int64)
        self.edges = edges
        self.queue_log = np.zeros((n, max_queue), dtype=np.int64)
        self.steps_done = 0
        self.n = n

    def update(self, queue_size, node, edge):
        self.income[node % self.n] += 1
        if edge > -1:
            self.traffic[self.edges[edge]] += 1
        for i in range(self.n):
            q = queue_size[i] + queue_size[i + self.n]
            self.queue_log[i, q] += 1
        self.steps_done += 1


def repr_logs(logs):
    """
    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Logs to display. Relies on `print`.

    Returns
    -------
    None
    """
    print(
        f"Arrivals: {logs.income.astype(int)}\n"
        f"Traffic: {logs.traffic.astype(int)}\n"
        f"Queues: {logs.queue_log.astype(int)}\n"
        f"Steps done: {logs.steps_done}"
    )



import numpy as np

from stochastic_matching.model import Model


def path_adjacency(n=2):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.

    Returns
    -------
    :class:`~numpy.ndarray`
        Adjacency matrix of a path graph :math:`P_n` (cf https://mathworld.wolfram.com/PathGraph.html).
    """
    adja = np.zeros([n, n], dtype=int)
    for i in range(n - 1):
        adja[i, i + 1] = 1
        adja[i + 1, i] = 1
    return adja


class Path(Model):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    Default is a two nodes line:

    >>> p2 = Path()
    >>> p2.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1],
           [1, 0]])

    Class name:

    >>> p2.name
    'Path'

    A five nodes line with alphabetical names:

    >>> p5 = Path(5, names="alpha")
    >>> p5.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 0, 0],
           [1, 0, 1, 0, 0],
           [0, 1, 0, 1, 0],
           [0, 0, 1, 0, 1],
           [0, 0, 0, 1, 0]])

    >>> [p5.names[i] for i in range(p5.n)]
    ['A', 'B', 'C', 'D', 'E']
    """

    name = "Path"

    def __init__(self, n=2, *args, **kwargs):
        adja = path_adjacency(n)
        super(Path, self).__init__(adjacency=adja, *args, **kwargs)


def star_adjacency(n=4):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.

    Returns
    -------
    :class:`~numpy.ndarray`
        Adjacency matrix of a star graph :math:`S_n` (https://mathworld.wolfram.com/StarGraph.html).
    """
    adja = np.zeros([n, n], dtype=int)
    for i in range(n - 1):
        adja[0, i + 1] = 1
        adja[i + 1, 0] = 1
    return adja


class Star(Model):
    """
    Parameters
    ----------

    n: :class:`int`
        Number of nodes.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    Default is a 4 nodes star (one center and three branches):

    >>> s4 = Star()
    >>> s4.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1],
           [1, 0, 0, 0],
           [1, 0, 0, 0],
           [1, 0, 0, 0]])

    A five nodes star:

    >>> s5 = Star(5)
    >>> s5.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1, 1],
           [1, 0, 0, 0, 0],
           [1, 0, 0, 0, 0],
           [1, 0, 0, 0, 0],
           [1, 0, 0, 0, 0]])
    """

    name = "Star"

    def __init__(self, n=4, *args, **kwargs):
        adja = star_adjacency(n)
        super(Star, self).__init__(adjacency=adja, *args, **kwargs)


def cycle_adjacency(n=3):
    """
    Parameters
    ----------
    n: :class:`int`
        Length of the cycle.

    Returns
    -------
    :class:`~numpy.ndarray`
        Adjacency matrix of a cycle graph :math:`C_n` (cf https://mathworld.wolfram.com/CycleGraph.html).
    """
    adja = path_adjacency(n)
    adja[0, -1] = 1
    adja[-1, 0] = 1
    return adja


class Cycle(Model):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    A triangle:

    >>> triangle = Cycle()
    >>> triangle.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1],
           [1, 0, 1],
           [1, 1, 0]])

    A pentacle:

    >>> pentacle = Cycle(n=5)
    >>> pentacle.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 0, 1],
           [1, 0, 1, 0, 0],
           [0, 1, 0, 1, 0],
           [0, 0, 1, 0, 1],
           [1, 0, 0, 1, 0]])

    A square:

    >>> square = Cycle(4)
    >>> square.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 1],
           [1, 0, 1, 0],
           [0, 1, 0, 1],
           [1, 0, 1, 0]])
    """

    name = "Cycle"

    def __init__(self, n=3, *args, **kwargs):
        adja = cycle_adjacency(n)
        super(Cycle, self).__init__(adjacency=adja, *args, **kwargs)


def complete_adjacency(n=3):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.

    Returns
    -------
    :class:`~numpy.ndarray`
        Adjacency matrix of a complete graph :math:`K_n` (cf https://mathworld.wolfram.com/CompleteGraph.html).
    """
    return np.ones([n, n], dtype=int) - np.identity(n, dtype=int)


class Complete(Model):
    """
    Parameters
    ----------
    n: :class:`int`
        Length of the cycle.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    A triangle:

    >>> triangle = Complete()
    >>> triangle.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1],
           [1, 0, 1],
           [1, 1, 0]])

    :math:`K_5`:

    >>> k5 = Complete(n=5)
    >>> k5.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1, 1],
           [1, 0, 1, 1, 1],
           [1, 1, 0, 1, 1],
           [1, 1, 1, 0, 1],
           [1, 1, 1, 1, 0]])

    :math:`K_4`:

    >>> k4 = Complete(4)
    >>> k4.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1],
           [1, 0, 1, 1],
           [1, 1, 0, 1],
           [1, 1, 1, 0]])
    """

    name = "Complete"

    def __init__(self, n=3, *args, **kwargs):
        adja = complete_adjacency(n)
        super(Complete, self).__init__(adjacency=adja, *args, **kwargs)


def concatenate_adjacency(adja_list, overlap=None):
    """
    Parameters
    ----------
    adja_list: :class:`list` of :class:`~numpy.ndarray`
        The adjacency matrices that one wants to merge.
    overlap: :class:`int` or :class:`list` of :class:`int`, optional
        Number of nodes that are common to two consecutive graphs. Default to 1.

    Returns
    -------
    :class:`~numpy.ndarray`
        Concatenated adjacency.

    Examples
    --------

    A codomino adjacency:

    >>> concatenate_adjacency([cycle_adjacency(), cycle_adjacency(4), cycle_adjacency()], 2) # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0, 0],
           [1, 0, 1, 0, 1, 0],
           [1, 1, 0, 1, 0, 0],
           [0, 0, 1, 0, 1, 1],
           [0, 1, 0, 1, 0, 1],
           [0, 0, 0, 1, 1, 0]])
    """
    adja_list = [a for a in adja_list if a.shape[0] > 0]
    na = len(adja_list)
    if overlap is None:
        overlap = 1
    if type(overlap) == int:
        overlap = [overlap] * (na - 1)
    n = sum(adja.shape[0] for adja in adja_list) - sum(overlap)
    adja = np.zeros([n, n], dtype=int)
    c_a = adja_list[0]
    c_n = c_a.shape[0]
    adja[:c_n, :c_n] = c_a
    pointer = c_n
    for i, o in enumerate(overlap):
        pointer -= o
        c_a = adja_list[i + 1]
        c_n = c_a.shape[0]
        adja[pointer : pointer + c_n, pointer : pointer + c_n] = c_a
        pointer += c_n
    return adja


def concatenate(model_list, overlap=None, *args, **kwargs):
    """
    Parameters
    ----------
    model_list: :class:`list` of :class:`~stochastic_matching.model.Model`
        The graphs that one want to merge. They must all be simple (and in particular have an adjacency matrix).
    overlap: :class:`int` or :class:`list` of :class:`int`, optional
        Number of nodes that are common to two consecutive graphs. Default to 1.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Returns
    -------
    :class:`~stochastic_matching.model.Model`
        Model of the concatenated graph.

    Examples
    --------

    A codomino graph:

    >>> codomino = concatenate([Cycle(), Cycle(4), Cycle()], 2)
    >>> codomino.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0, 0],
           [1, 0, 1, 0, 1, 0],
           [1, 1, 0, 1, 0, 0],
           [0, 0, 1, 0, 1, 1],
           [0, 1, 0, 1, 0, 1],
           [0, 0, 0, 1, 1, 0]])
    """
    adja_list = [g.adjacency for g in model_list]
    adja = concatenate_adjacency(adja_list, overlap)
    return Model(adjacency=adja, *args, **kwargs)


class Codomino(Model):
    """
    Parameters
    ----------
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------
    >>> codomino = Codomino()
    >>> codomino.adjacency
    array([[0, 1, 1, 0, 0, 0],
           [1, 0, 1, 0, 1, 0],
           [1, 1, 0, 1, 0, 0],
           [0, 0, 1, 0, 1, 1],
           [0, 1, 0, 1, 0, 1],
           [0, 0, 0, 1, 1, 0]])
    """

    name = "Codomino"

    def __init__(self, *args, **kwargs):
        adja = concatenate_adjacency(
            [cycle_adjacency(), cycle_adjacency(4), cycle_adjacency()], 2
        )
        super(Codomino, self).__init__(adjacency=adja, *args, **kwargs)


class ErdosRenyi(Model):
    """
    Parameters
    ----------
    n: :class:`int`
        Number of nodes.
    d: :class:`int` or :class:`float`
        Target degree of Erdos Renyi graph.
    seed: :class:`int`, optional
        Random number generator seed.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    >>> erdos_renyi = ErdosRenyi(n=10, d=3, seed=42)
    >>> erdos_renyi.m
    20
    >>> erdos_renyi.adjacency.astype(int) # doctest: +NORMALIZE_WHITESPACE
    array([[0, 0, 0, 0, 1, 1, 1, 0, 0, 0],
           [0, 0, 0, 1, 1, 1, 1, 0, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
           [0, 1, 0, 0, 0, 0, 1, 1, 0, 0],
           [1, 1, 0, 0, 0, 0, 1, 0, 0, 1],
           [1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 0, 0, 1, 0],
           [0, 0, 0, 1, 0, 1, 0, 0, 0, 1],
           [0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
           [0, 1, 1, 0, 1, 1, 0, 1, 0, 0]])
    """
    name = "Erdös-Rényi"

    def __init__(self, n=10, d=3, seed=None, *args, **kwargs):
        if seed is not None:
            np.random.seed(seed)
        p = d / (n - 1)
        edges = np.triu(np.random.rand(n, n) < p, 1)
        adjacency = edges + edges.T
        super(ErdosRenyi, self).__init__(adjacency=adjacency, *args, **kwargs)


class NS19(Model):
    """
    Parameters
    ----------
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------
    >>> ns = NS19()
    >>> ns.incidence
    array([[1, 0, 0, 0, 1, 0, 0],
           [0, 1, 0, 0, 1, 1, 1],
           [0, 0, 1, 0, 0, 1, 1],
           [0, 0, 0, 1, 0, 0, 1]])
    """
    name = "Nazari-Stolyar"

    def __init__(self, *args, **kwargs):
        incidence = [
            [1, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 1, 1, 1],
            [0, 0, 1, 0, 0, 1, 1],
            [0, 0, 0, 1, 0, 0, 1],
        ]
        super(NS19, self).__init__(incidence=incidence, *args, **kwargs)


class Pyramid(Model):
    """
    Parameters
    ----------
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------
    >>> pyramid = Pyramid()
    >>> pyramid.adjacency
    array([[0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 1, 0, 0, 0, 0],
           [1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
           [0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 1, 0, 1, 0, 0, 1, 0],
           [0, 1, 0, 0, 1, 0, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 0, 1, 1],
           [0, 0, 0, 0, 1, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 0, 1, 1, 0]])
    """

    name = "Pyramid"

    def __init__(self, *args, **kwargs):
        adja = concatenate_adjacency(
            [
                cycle_adjacency(),
                cycle_adjacency(5),
                cycle_adjacency(5),
                cycle_adjacency(),
            ],
            2,
        )
        super(Pyramid, self).__init__(adjacency=adja, *args, **kwargs)


class Tadpole(Model):
    """
    Parameters
    ----------
    m: :class:`int`
        Length of the cycle.
    n: :class:`int`
        Length of the tail.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.


    Examples
    --------

    A triangle with a one-edge tail (paw graph):

    >>> paw = Tadpole()
    >>> paw.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0],
           [1, 0, 1, 0],
           [1, 1, 0, 1],
           [0, 0, 1, 0]])

    A pentacle:

    >>> c5 = Tadpole(m=5, n=0)
    >>> c5.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 0, 1],
           [1, 0, 1, 0, 0],
           [0, 1, 0, 1, 0],
           [0, 0, 1, 0, 1],
           [1, 0, 0, 1, 0]])

    A larger tadpole:

    >>> long_pan = Tadpole(m=4, n=3)
    >>> long_pan.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 1, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0],
           [0, 1, 0, 1, 0, 0, 0],
           [1, 0, 1, 0, 1, 0, 0],
           [0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 1, 0]])
    """

    name = "Tadpole"

    def __init__(self, m=3, n=1, *args, **kwargs):
        adja = concatenate_adjacency(
            adja_list=[cycle_adjacency(m), path_adjacency(n + 1)], overlap=1
        )
        super(Tadpole, self).__init__(adjacency=adja, *args, **kwargs)


class Lollipop(Model):
    """
    Parameters
    ----------
    m: :class:`int`
        Length of the complete graph.
    n: :class:`int`
        Length of the tail.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    A triangle with a one-edge tail (paw graph):

    >>> l_3_1 = Lollipop()
    >>> l_3_1.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0],
           [1, 0, 1, 0],
           [1, 1, 0, 1],
           [0, 0, 1, 0]])


    A larger lollipop:

    >>> l_5_3 = Lollipop(m=5, n=3)
    >>> l_5_3.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1, 1, 0, 0, 0],
           [1, 0, 1, 1, 1, 0, 0, 0],
           [1, 1, 0, 1, 1, 0, 0, 0],
           [1, 1, 1, 0, 1, 0, 0, 0],
           [1, 1, 1, 1, 0, 1, 0, 0],
           [0, 0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 0]])
    """

    name = "Lollipop"

    def __init__(self, m=3, n=1, *args, **kwargs):
        adja = concatenate_adjacency(
            adja_list=[complete_adjacency(m), path_adjacency(n + 1)], overlap=1
        )
        super(Lollipop, self).__init__(adjacency=adja, *args, **kwargs)


class KayakPaddle(Model):
    """
    Parameters
    ----------
    k: :class:`int`
        Size of the first cycle.
    l: :class:`int`
        Length of the path that joins the two cycles.
    m: :class:`int`, optional
        Size of the second cycle. Default to the size of the first cycle.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    A square and a triangle joined by a path of length 3.

    >>> kayak = KayakPaddle(k=4, m=3, l=3)
    >>> kayak.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 1, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0, 0],
           [0, 1, 0, 1, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 0, 1, 0, 1, 0, 0, 0],
           [0, 0, 0, 0, 1, 0, 1, 0, 0],
           [0, 0, 0, 0, 0, 1, 0, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 1, 0]])

    A bow-tie (two triangles with one node in common).

    >>> KayakPaddle(l=0).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0],
           [1, 0, 1, 0, 0],
           [1, 1, 0, 1, 1],
           [0, 0, 1, 0, 1],
           [0, 0, 1, 1, 0]])
    """

    name = "Kayak Paddle"

    def __init__(self, k=3, l=1, m=None, *args, **kwargs):
        if m is None:
            m = k
        adja = concatenate_adjacency(
            [cycle_adjacency(k), path_adjacency(l + 1), cycle_adjacency(m)]
        )
        super(KayakPaddle, self).__init__(adjacency=adja, *args, **kwargs)


class Barbell(Model):
    """
    Parameters
    ----------
    k: :class:`int`
        Size of the first complete graph.
    m: :class:`int`, optional
        Size of the second complete graph. Default to the size of the first complete graph.
    l: :class:`int`
        Length of the path that joins the two complete graphs.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    Traditional barbel graph with complete graphs of same size and unit bridge.

    >>> barbel_5 = Barbell(5)
    >>> barbel_5.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
           [1, 0, 1, 1, 1, 0, 0, 0, 0, 0],
           [1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
           [1, 1, 1, 0, 1, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 0, 1, 1, 1, 1],
           [0, 0, 0, 0, 0, 1, 0, 1, 1, 1],
           [0, 0, 0, 0, 0, 1, 1, 0, 1, 1],
           [0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
           [0, 0, 0, 0, 0, 1, 1, 1, 1, 0]])

    A bow-tie (two triangles with one node in common).

    >>> Barbell(l=0).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0],
           [1, 0, 1, 0, 0],
           [1, 1, 0, 1, 1],
           [0, 0, 1, 0, 1],
           [0, 0, 1, 1, 0]])

    Something more elaborated.

    >>> Barbell(k=3, m=5, l=4).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 1, 0, 1, 1, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0]])
    """

    name = "Barbell"

    def __init__(self, k=3, l=1, m=None, *args, **kwargs):
        if m is None:
            m = k
        adja = concatenate_adjacency(
            [complete_adjacency(k), path_adjacency(l + 1), complete_adjacency(m)]
        )
        super(Barbell, self).__init__(adjacency=adja, *args, **kwargs)


class CycleChain(Model):
    """
    Parameters
    ----------
    n: :class:`int`
        Size of the cycles.
    c: :class:`int`
        Number of copies.
    o: :class:`int`
        Overlap between cycles.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    The diamond graph (two triangles).

    >>> diamond = CycleChain()
    >>> diamond.adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0],
           [1, 0, 1, 1],
           [1, 1, 0, 1],
           [0, 1, 1, 0]])

    The *triamond* graph.

    >>> CycleChain(n=3, c=3).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0],
           [1, 0, 1, 1, 0],
           [1, 1, 0, 1, 1],
           [0, 1, 1, 0, 1],
           [0, 0, 1, 1, 0]])

    The triangular snake graph :math:`TS_9` (cf https://mathworld.wolfram.com/TriangularSnakeGraph.html)

    >>> CycleChain(n=3, c=4, o=1).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 1, 0, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0, 0],
           [1, 1, 0, 1, 1, 0, 0, 0, 0],
           [0, 0, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 1, 1, 0, 1, 1, 0, 0],
           [0, 0, 0, 0, 1, 0, 1, 0, 0],
           [0, 0, 0, 0, 1, 1, 0, 1, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 1, 0]])

    The domino graph, or 3-ladder graph (cf https://mathworld.wolfram.com/LadderGraph.html)

    >>> CycleChain(n=4, c=2).adjacency # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 1, 0, 0],
           [1, 0, 1, 0, 0, 0],
           [0, 1, 0, 1, 0, 1],
           [1, 0, 1, 0, 1, 0],
           [0, 0, 0, 1, 0, 1],
           [0, 0, 1, 0, 1, 0]])
    """

    name = "Cycle Chain"

    def __init__(self, n=3, c=2, o=2, *args, **kwargs):
        adja = concatenate_adjacency([cycle_adjacency(n)] * c, o)
        super(CycleChain, self).__init__(adjacency=adja, *args, **kwargs)


class HyperPaddle(Model):
    """
    Parameters
    ----------
    k: :class:`int`
        Size of the first cycle.
    m: :class:`int`
        Size of the second cycle.
    l: :class:`int`
        Length of the path of 3-edges that joins the two cycles.
    *args
        Positional parameters for the model.
    **kwargs: :class:`dict`, optional
        Keyword parameters for the model.

    Examples
    --------

    The *candy*, a basic but very useful hypergraph.

    >>> candy = HyperPaddle()
    >>> candy.incidence
    array([[1, 1, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0],
           [0, 1, 1, 0, 0, 0, 1],
           [0, 0, 0, 0, 0, 0, 1],
           [0, 0, 0, 1, 1, 0, 1],
           [0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 1, 1, 0]])

    A more elaborate hypergraph

    >>> HyperPaddle(k=5, m=4, l=3).incidence
    array([[1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
           [0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
           [0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1],
           [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0]])

    Warning: without any hyperedge, we have two disconnected cycles.

    >>> HyperPaddle(l=0).incidence
    array([[1, 1, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0],
           [0, 1, 1, 0, 0, 0],
           [0, 0, 0, 1, 1, 0],
           [0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 1, 1]])
    """

    name = "Hyper Kayak Paddle"

    def __init__(self, k=3, m=3, l=1, *args, **kwargs):
        n = k + m + l
        incidence = np.zeros((n, n), dtype=int)
        left = Cycle(n=k).incidence
        incidence[:k, :k] = left
        right = Cycle(n=m).incidence
        incidence[(n - m) :, k : (k + m)] = right
        for i in range(l):
            incidence[(k - 1 + i) : (k + 2 + i), n - l + i] = 1
        super(HyperPaddle, self).__init__(incidence=incidence, *args, **kwargs)


class Fan(Model):
    """
    Parameters
    ----------
    cycles: :class:`int`
        Number of cycles
    cycle_size: :class:`int`
        Size of cycles
    hyperedges: :class:`int`
        Number of hyperedges that connect the cycles.
    *args
        Positional parameters for the model.
    **kwargs
        Keyword parameters for the model.

    Examples
    --------

    A basic 3-leaves clover:

    >>> clover = Fan()
    >>> clover.incidence  # doctest: +NORMALIZE_WHITESPACE
    array([[1, 1, 0, 0, 0, 0, 0, 0, 0, 1],
           [1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
           [0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 1, 1, 0, 0, 0, 0, 1],
           [0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 1, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 0, 0, 0, 1, 1, 0]])

    Increase the hyperedge connectivity:

    >>> connected = Fan(hyperedges=2)
    >>> connected.incidence  # doctest: +NORMALIZE_WHITESPACE
    array([[1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
           [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1],
           [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0],
           [0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1],
           [0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0],
           [0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1],
           [0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]])

    With only two cycles, we have a simple graph.

    >>> db = Fan(cycles=2, cycle_size=4)
    >>> db.incidence # doctest: +NORMALIZE_WHITESPACE
    array([[1, 1, 0, 0, 0, 0, 0, 0, 1],
           [1, 0, 1, 0, 0, 0, 0, 0, 0],
           [0, 0, 1, 1, 0, 0, 0, 0, 0],
           [0, 1, 0, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 1, 0, 0, 1],
           [0, 0, 0, 0, 1, 0, 1, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 1, 0],
           [0, 0, 0, 0, 0, 1, 0, 1, 0]])
    >>> from stochastic_matching.model import incidence_to_adjacency
    >>> incidence_to_adjacency(db.incidence) # doctest: +NORMALIZE_WHITESPACE
    array([[0, 1, 0, 1, 1, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0],
           [0, 1, 0, 1, 0, 0, 0, 0],
           [1, 0, 1, 0, 0, 0, 0, 0],
           [1, 0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 1, 0, 1, 0],
           [0, 0, 0, 0, 0, 1, 0, 1],
           [0, 0, 0, 0, 1, 0, 1, 0]])
    """

    name = "Fan"

    def __init__(self, cycles=3, cycle_size=3, hyperedges=1, *args, **kwargs):
        n = cycles * cycle_size
        incidence = np.zeros((n, n + hyperedges), dtype=int)
        cycle_incidence = Cycle(n=cycle_size).incidence
        for c in range(cycles):
            incidence[
                (c * cycle_size) : ((c + 1) * cycle_size),
                (c * cycle_size) : ((c + 1) * cycle_size),
            ] = cycle_incidence
            for h in range(hyperedges):
                incidence[c * cycle_size + h, h - hyperedges] = 1
        super(Fan, self).__init__(incidence=incidence, *args, **kwargs)



import numpy as np
from numba import njit
from matplotlib import pyplot as plt

from stochastic_matching.simulator.arrivals import Arrivals
from stochastic_matching.simulator.graph import make_jit_graph
from stochastic_matching.simulator.logs import Logs
from stochastic_matching.display import int_2_str
from stochastic_matching.simulator.logs import repr_logs
from stochastic_matching.simulator.metrics import AvgQueues, Regret, Delay, CCDF, Flow


@njit
def core_simulator(logs, arrivals, graph, n_steps, queue_size, selector):
    """
    Generic jitted function for queue-size based policies.

    Parameters
    ----------
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables.
    arrivals: :class:`~stochastic_matching.simulator.arrivals.Arrivals`
        Item arrival process.
    graph: :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Model graph.
    n_steps: :class:`int`
        Number of arrivals to process.
    queue_size: :class:`~numpy.ndarray`
        Number of waiting items of each class.
    selector: callable
        Jitted function that selects edge (or not) based on graph, queue_size, and arriving node.

    Returns
    -------
    None
    """

    n, max_queue = logs.queue_log.shape

    for age in range(n_steps):
        node = arrivals.draw()
        queue_size[node] += 1
        if queue_size[node] == max_queue:
            return None

        best_edge = selector(graph=graph, queue_size=queue_size, node=node)

        if best_edge > -1:
            queue_size[graph.nodes(best_edge)] -= 1

        logs.update(queue_size=queue_size, node=node, edge=best_edge)


class Simulator:
    """
    Abstract class that describes the generic behavior of matching simulators.
    See subclasses for detailed examples.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    n_steps: :class:`int`, optional
        Number of arrivals to simulate.
    seed: :class:`int`, optional
        Seed of the random generator
    max_queue: :class:`int`
        Max queue size. Necessary for speed and detection of unstability.
        For stable systems very close to the unstability
        border, the max_queue may be reached.

    Attributes
    ----------
    internal: :class:`dict`
        Inner variables. Default to arrivals, graphs, queue_size, n_steps. Subclasses can add other variables.
    logs: :class:`~stochastic_matching.simulator.logs.Logs`
        Monitored variables (traffic on edges, queue size distribution, number of steps achieved).

    Examples
    --------

    >>> import stochastic_matching as sm
    >>> sim = sm.FCFM(sm.CycleChain(rates=[2, 2.1, 1.1, 1]), seed=42, n_steps=1000, max_queue=8)
    >>> sim
    Simulator of type fcfm.

    Use :meth:`~stochastic_matching.simulator.simulator.Simulator.run` to make the simulation.

    >>> sim.run()

    Raw results are stored in `logs`. A `plog` property gives a pretty print of the logs.

    >>> sim.plogs #doctest: +NORMALIZE_WHITESPACE
    Arrivals: [67 80 43 39]
    Traffic: [43 17 14 23 12]
    Queues: [[118  47  26  15  14   7   1   1]
     [188  25  13   3   0   0   0   0]
     [217   8   3   1   0   0   0   0]
     [125  50  31  11   9   3   0   0]]
    Steps done: 229

    Different methods are proposed to provide various indicators.

    >>> sim.avg_queues
    array([1.08296943, 0.26200873, 0.07423581, 0.8558952 ])

    >>> sim.delay
    np.float64(0.3669530919847866)

    >>> sim.ccdf #doctest: +NORMALIZE_WHITESPACE
    array([[1.        , 0.48471616, 0.27947598, 0.16593886, 0.10043668,
            0.03930131, 0.00873362, 0.00436681, 0.        ],
           [1.        , 0.1790393 , 0.069869  , 0.01310044, 0.        ,
            0.        , 0.        , 0.        , 0.        ],
           [1.        , 0.05240175, 0.01746725, 0.00436681, 0.        ,
            0.        , 0.        , 0.        , 0.        ],
           [1.        , 0.45414847, 0.23580786, 0.10043668, 0.05240175,
            0.01310044, 0.        , 0.        , 0.        ]])
    >>> sim.flow
    array([1.16419214, 0.46026201, 0.3790393 , 0.62270742, 0.32489083])

    You can also draw the average or CCDF of the queues.

    >>> fig = sim.show_average_queues()
    >>> fig #doctest: +ELLIPSIS
    <Figure size ...x... with 1 Axes>

    >>> fig = sim.show_average_queues(indices=[0, 3, 2], sort=True, as_time=True)
    >>> fig #doctest: +ELLIPSIS
    <Figure size ...x... with 1 Axes>

    >>> fig = sim.show_ccdf()
    >>> fig #doctest: +ELLIPSIS
    <Figure size ...x... with 1 Axes>

    >>> fig = sim.show_ccdf(indices=[0, 3, 2], sort=True, strict=True)
    >>> fig #doctest: +ELLIPSIS
    <Figure size ...x... with 1 Axes>
    """

    name = None
    """
    Name that can be used to list all non-abstract classes.
    """

    def __init__(self, model, n_steps=1000000, seed=None, max_queue=1000):
        self.model = model
        self.max_queue = max_queue
        self.n_steps = n_steps
        self.seed = seed

        self.internal = None
        self.set_internal()

        self.logs = Logs(n=self.model.n, m=self.model.m, max_queue=self.max_queue)

    @property
    def plogs(self):
        """
        Print logs.

        Returns
        -------
        None
        """
        repr_logs(self.logs)

    def set_internal(self):
        """
        Populate the internal state.

        Returns
        -------
        None
        """
        self.internal = {
            "arrivals": Arrivals(mu=self.model.rates, seed=self.seed),
            "graph": make_jit_graph(self.model),
            "n_steps": self.n_steps,
            "queue_size": np.zeros(self.model.n, dtype=int),
        }

    def reset(self):
        """
        Reset internal state and monitored variables.

        Returns
        -------
        None
        """
        self.set_internal()
        self.logs = Logs(n=self.model.n, m=self.model.m, max_queue=self.max_queue)

    def run(self):
        """
        Run simulation.
        Results are stored in the attribute :attr:`~stochastic_matching.simulator.simulator.Simulator.logs`.

        Returns
        -------
        None
        """
        core_simulator(logs=self.logs, **self.internal)

    @property
    def avg_queues(self):
        """
        :class:`~numpy.ndarray`
            Average queue sizes.
        """
        return AvgQueues.get(self)

    @property
    def delay(self):
        """
        :class:`float`
            Average waiting time
        """
        return Delay.get(self)

    @property
    def ccdf(self):
        """
        :class:`~numpy.ndarray`
            CCDFs of the queues.
        """
        return CCDF.get(self)

    @property
    def flow(self):
        """
        Normalize the simulated flow.

        Returns
        -------
        :class:`~numpy.ndarray`
            Flow on edges.
        """
        return Flow.get(self)

    @property
    def regret(self):
        return Regret.get(self)

    def show_average_queues(self, indices=None, sort=False, as_time=False):
        """
        Parameters
        ----------
        indices: :class:`list`, optional
            Indices of the nodes to display
        sort: :class:`bool`, optional
            If True, display the nodes by decreasing average queue size
        as_time: :class:`bool`, optional
            If True, display the nodes by decreasing average queue size

        Returns
        -------
        :class:`~matplotlib.figure.Figure`
            A figure of the CCDFs of the queues.
        """
        averages = self.avg_queues
        if as_time:
            averages = averages / self.model.rates
        if indices is not None:
            averages = averages[indices]
            names = [int_2_str(self.model, i) for i in indices]
        else:
            names = [int_2_str(self.model, i) for i in range(self.model.n)]
        if sort is True:
            ind = np.argsort(-averages)
            averages = averages[ind]
            names = [names[i] for i in ind]
        plt.bar(names, averages)
        if as_time:
            plt.ylabel("Average waiting time")
        else:
            plt.ylabel("Average queue occupancy")
        plt.xlabel("Node")
        return plt.gcf()

    def show_ccdf(self, indices=None, sort=None, strict=False):
        """
        Parameters
        ----------
        indices: :class:`list`, optional
            Indices of the nodes to display
        sort: :class:`bool`, optional
            If True, order the nodes by decreasing average queue size
        strict: :class:`bool`, default = False
            Draws the curves as a true piece-wise function

        Returns
        -------
        :class:`~matplotlib.figure.Figure`
            A figure of the CCDFs of the queues.
        """
        ccdf = self.ccdf

        if indices is not None:
            ccdf = ccdf[indices, :]
            names = [int_2_str(self.model, i) for i in indices]
        else:
            names = [int_2_str(self.model, i) for i in range(self.model.n)]
        if sort is True:
            averages = self.avg_queues
            if indices is not None:
                averages = averages[indices]
            ind = np.argsort(-averages)
            ccdf = ccdf[ind, :]
            names = [names[i] for i in ind]
        for i, name in enumerate(names):
            if strict:
                data = ccdf[i, ccdf[i, :] > 0]
                n_d = len(data)
                x = np.zeros(2 * n_d - 1)
                x[::2] = np.arange(n_d)
                x[1::2] = np.arange(n_d - 1)
                y = np.zeros(2 * n_d - 1)
                y[::2] = data
                y[1::2] = data[1:]
                plt.semilogy(x, y, label=name)
            else:
                plt.semilogy(ccdf[i, ccdf[i, :] > 0], label=name)
        plt.legend()
        plt.xlim([0, None])
        plt.ylim([None, 1])
        plt.ylabel("CCDF")
        plt.xlabel("Queue occupancy")
        return plt.gcf()

    def __repr__(self):
        return f"Simulator of type {self.name}."



from numba import int32
from numba.experimental import jitclass


def make_jit_graph(model):
    """
    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model that contains an incidence matrix.

    Returns
    -------
    :class:`~stochastic_matching.simulator.graph.JitHyperGraph`
        Jitted graph.
    """
    return JitHyperGraph(
        incid_ind=model.incidence_csr.indices,
        incid_ptr=model.incidence_csr.indptr,
        coinc_ind=model.incidence_csc.indices,
        coinc_ptr=model.incidence_csc.indptr,
        n=model.n,
        m=model.m,
    )


@jitclass
class JitHyperGraph:
    """
    Jit compatible view of a (hyper)graph.

    Parameters
    ----------
    incid_ind: :class:`~numpy.ndarray`
        Indices of the incidence matrix.
    incid_ptr: :class:`~numpy.ndarray`
        Pointers of the incidence matrix.
    coinc_ind: :class:`~numpy.ndarray`
        Indices of the co-incidence matrix.
    coinc_ptr: :class:`~numpy.ndarray`
        Pointers of the co-incidence matrix.
    n: :class:`int`
        Number of nodes.
    m: :class:`int`
        Number of (hyper)edges.
    """

    incid_ind: int32[:]
    incid_ptr: int32[:]
    coinc_ind: int32[:]
    coinc_ptr: int32[:]
    n: int
    m: int

    def __init__(self, incid_ind, incid_ptr, coinc_ind, coinc_ptr, n, m):
        self.incid_ind = incid_ind
        self.incid_ptr = incid_ptr
        self.coinc_ind = coinc_ind
        self.coinc_ptr = coinc_ptr
        self.n = n
        self.m = m

    def edges(self, node):
        """
        Parameters
        ----------
        node: :class:`int`
            Index of a node.

        Returns
        -------
        :class:`~numpy.ndarray`
            Edges adjacent to the node.
        """
        return self.incid_ind[self.incid_ptr[node] : self.incid_ptr[node + 1]]

    def nodes(self, edge):
        """
        Parameters
        ----------
        edge: :class:`int`
            Index of an edge.

        Returns
        -------
        :class:`~numpy.ndarray`
            Nodes adjacent to the edge.
        """
        return self.coinc_ind[self.coinc_ptr[edge] : self.coinc_ptr[edge + 1]]



from numba.experimental import jitclass
from numba import typeof, int64
from numba.typed import Dict
import numpy as np
import os


if os.environ.get("NUMBA_DISABLE_JIT") == "1":
    vq_type = dict
else:
    vq_type = typeof(Dict.empty(key_type=int64, value_type=int64))

infinity = 2**63 - 1
infinity_type = typeof(infinity)


@jitclass
class MultiQueue:
    """
    A cruder, faster implementation of FullMultiQueue. Queues must be explicitly bounded on initialisation.

    Parameters
    ----------
    n: :class:`int`
        Number of classes.

    Examples
    --------

    Let's populate a multiqueue and play with it.

    >>> queues = MultiQueue(4, max_queue=4)
    >>> _ = [queues.add(i, s) for i, s in [(0, 2), (0, 5), (1, 4), (3, 0), (3, 1), (3, 6)]]
    >>> [int(queues.size(i)) for i in range(4)]
    [2, 1, 0, 3]

    The age of the oldest element of an empty queue is infinity.

    >>> [int(queues.oldest(i)) for i in range(4)]
    [2, 4, 9223372036854775807, 0]
    >>> queues.add(3, 10)
    >>> [int(queues.size(i)) for i in range(4)]
    [2, 1, 0, 4]

    Trying to go above the max queue raises an error (item is not added, queue stays viable):

    >>> queues.add(3, 10)
    Traceback (most recent call last):
    ...
    OverflowError: Max queue size reached.

    So does trying to pop something empty (and queue stays viable).

    >>> queues.pop(2)
    Traceback (most recent call last):
    ...
    ValueError: Try to pop an empty queue

    Let's play a bit more.

    >>> _ = [queues.pop(i) for i in [0, 1, 3, 3]]
    >>> [int(queues.size(i)) for i in range(4)]
    [1, 0, 0, 2]
    >>> [int(queues.oldest(i)) for i in range(4)]
    [5, 9223372036854775807, 9223372036854775807, 6]
    >>> for a in range(22, 30):
    ...     print(f"Add {a} to class 3.")
    ...     queues.add(3, a)
    ...     print(f"Oldest of class 3 is {queues.oldest(3)}, we pop it.")
    ...     queues.pop(3)
    Add 22 to class 3.
    Oldest of class 3 is 6, we pop it.
    Add 23 to class 3.
    Oldest of class 3 is 10, we pop it.
    Add 24 to class 3.
    Oldest of class 3 is 22, we pop it.
    Add 25 to class 3.
    Oldest of class 3 is 23, we pop it.
    Add 26 to class 3.
    Oldest of class 3 is 24, we pop it.
    Add 27 to class 3.
    Oldest of class 3 is 25, we pop it.
    Add 28 to class 3.
    Oldest of class 3 is 26, we pop it.
    Add 29 to class 3.
    Oldest of class 3 is 27, we pop it.
    >>> [int(queues.size(i)) for i in range(4)]
    [1, 0, 0, 2]
    >>> [int(queues.oldest(i)) for i in range(4)]
    [5, 9223372036854775807, 9223372036854775807, 28]
    """

    queues: int64[:, :]
    heads: int64[:]
    tails: int64[:]
    infinity: infinity_type
    max_queue: infinity_type

    def __init__(self, n, max_queue=10000):
        self.infinity = 2**63 - 1
        self.queues = np.ones((n, max_queue), dtype=np.int64)
        self.queues[:] = self.infinity
        self.heads = np.zeros(n, dtype=np.int64)
        self.tails = np.zeros(n, dtype=np.int64)
        self.max_queue = max_queue

    def add(self, item_class, stamp):
        """
        Parameters
        ----------
        item_class: :class:`int`
            Class of the item.
        stamp: :class:`int`
            Item stamp to add to the queue of the item class.

        Returns
        -------
        None

        """
        if self.queues[item_class, self.tails[item_class]] != self.infinity:
            raise OverflowError("Max queue size reached.")
        self.queues[item_class, self.tails[item_class]] = stamp
        self.tails[item_class] = (self.tails[item_class] + 1) % self.max_queue

    def oldest(self, item_class):
        """
        Parameters
        ----------
        item_class: :class:`int`
            Class of the item.

        Returns
        -------
        :class:`int`
            Stamp of the oldest item (or infinity if the queue is empty).
        """
        return self.queues[item_class, self.heads[item_class]]

    def pop(self, item_class):
        """
        Remove the oldest item from given class. Item stamp is NOT returned (use oldest beforehand if you need access).

        Parameters
        ----------
        item_class: :class:`int`
            Class of the item to remove.

        Returns
        -------
        None
        """

        if self.queues[item_class, self.heads[item_class]] == self.infinity:
            raise ValueError("Try to pop an empty queue")
        self.queues[item_class, self.heads[item_class]] = self.infinity
        self.heads[item_class] = (self.heads[item_class] + 1) % self.max_queue

    def size(self, item_class):
        if self.queues[item_class, self.tails[item_class]] == self.infinity:
            return (self.tails[item_class] - self.heads[item_class]) % self.max_queue
        return self.max_queue


@jitclass
class FullMultiQueue:
    """
    A simple management of multi-class FCFM that supports add to tail
    and pop from head over multiple item classes.
    Each item must be characterized by a UNIQUE non-negative int (typically a timestamp).

    Parameters
    ----------
    n: :class:`int`
        Number of classes.

    Examples
    --------

    Let's populate a multiqueue.

    >>> queues = FullMultiQueue(4)
    >>> for cl, stamp in [(0, 2), (0, 5), (1, 4), (3, 0), (3, 1), (3, 6)]:
    ...     queues.add(cl, stamp)

    Oldest items per class (infinity indicates an empty queue):

    >>> [queues.oldest(i) for i in range(4)]
    [2, 4, 9223372036854775807, 0]

    Let's remove some item. Note that pop returns nothing.

    >>> [queues.pop(i) for i in [0, 1, 3, 3]]
    [None, None, None, None]

    Oldest items per class (infinity indicates an empty queue):

    >>> [queues.oldest(i) for i in range(4)]
    [5, 9223372036854775807, 9223372036854775807, 6]

    Note that trying to pop from an empty queue will raise an error:

    >>> queues.pop(1)
    Traceback (most recent call last):
    ...
    KeyError: 9223372036854775807
    """

    tails: int64[:]
    vq: vq_type
    infinity: infinity_type

    def __init__(self, n):
        self.infinity = 2**63 - 1
        self.tails = np.array([-1 - e for e in range(n)], dtype=np.int64)
        self.vq = {t: self.infinity for t in self.tails}

    def add(self, item_class, stamp):
        """
        Parameters
        ----------
        item_class: :class:`int`
            Class of the item.
        stamp: :class:`int`
            Item stamp to add to the queue of the item class.

        Returns
        -------
        None

        """
        self.vq[self.tails[item_class]] = stamp  # tail points to a real age
        self.vq[stamp] = self.infinity  # new age points to infinity
        self.tails[item_class] = stamp  # Move up tail

    def oldest(self, item_class):
        """
        Parameters
        ----------
        item_class: :class:`int`
            Class of the item.

        Returns
        -------
        :class:`int`
            Stamp of the oldest item (or infinity if the queue is empty).
        """
        return self.vq[-1 - item_class]

    def pop(self, item_class):
        """
        Remove the oldest item from given class. Item stamp is NOT returned (use oldest beforehand if you need access).

        Parameters
        ----------
        item_class: :class:`int`
            Class of the item to remove.

        Returns
        -------
        None
        """
        stamp = self.oldest(item_class)  # get oldest item of item_class
        self.vq[-1 - item_class] = self.vq[stamp]  # head points to next item
        del self.vq[stamp]  # remove oldest item
        if stamp == self.tails[item_class]:  # if queue is empty, plug tail to head
            self.tails[item_class] = -1 - item_class



import numpy as np
from numba import njit, int64, float64
from numba.experimental import jitclass


@njit
def create_prob_alias(mu):
    """
    Prepare vector to draw a distribution with the alias method.

    Based on https://www.keithschwarz.com/darts-dice-coins/.

    Parameters
    ----------
    mu: :class:`list` or :class:`~numpy.ndarray`
        Arrival intensities.

    Returns
    -------
    prob: :class:`~numpy.ndarray`
        Probabilities to stay in the drawn bucket
    alias: :class:`~numpy.ndarray`
        Redirection array

    Examples
    --------

    >>> probas, aliases = create_prob_alias([2 ,2, 3, 1])
    >>> probas
    array([1. , 1. , 1. , 0.5])
    >>> aliases.astype(int)
    array([0, 0, 0, 2])
    """
    if isinstance(mu, list):
        cmu = np.array(mu)
    else:
        cmu = mu
    n = len(cmu)
    alias = np.zeros(n, dtype=np.int64)
    prob = np.zeros(n)
    # noinspection PyUnresolvedReferences
    normalized_intensities = cmu * n / np.sum(cmu)
    small = [i for i in range(n) if normalized_intensities[i] < 1]
    large = [i for i in range(n) if normalized_intensities[i] >= 1]
    while small and large:
        l, g = small.pop(), large.pop()
        prob[l], alias[l] = normalized_intensities[l], g
        normalized_intensities[g] += normalized_intensities[l] - 1
        if normalized_intensities[g] < 1:
            small.append(g)
        else:
            large.append(g)
    for i in large + small:
        prob[i] = 1
    return prob, alias


@jitclass
class Arrivals:
    """
    Parameters
    ----------
    mu: :class:`list` or :class:`~numpy.ndarray`
        Arrival intensities.
    seed: :class:`int`
        Seed for random generator.

    Examples
    --------

    >>> arrivals = Arrivals([2 ,2, 3, 1], seed=42)
    >>> arrivals.prob
    array([1. , 1. , 1. , 0.5])
    >>> arrivals.alias.astype(int)
    array([0, 0, 0, 2])
    >>> from collections import Counter
    >>> Counter([arrivals.draw() for _ in range(800)])
    Counter({2: 291, 1: 210, 0: 208, 3: 91})
    """

    prob: float64[:]
    alias: int64[:]
    n: int

    def __init__(self, mu, seed=None):
        self.prob, self.alias = create_prob_alias(mu)
        self.n = len(mu)
        if seed is not None:
            self.set_seed(seed)

    @staticmethod
    def set_seed(seed):
        np.random.seed(seed)

    def draw(self):
        node = np.random.randint(self.n)
        if np.random.rand() > self.prob[node]:
            node = self.alias[node]
        return node



import itertools
import numpy as np

from stochastic_matching.simulator.graph import make_jit_graph
from stochastic_matching.simulator.simulator import Simulator
from stochastic_matching.model import Model
from stochastic_matching.common import class_converter
from stochastic_matching.simulator.logs import PhantomLogs


def expand_model(model, forbidden_edges, epsilon):
    """
    Prepares a model for epsilon-filtering.

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Initial model.
    forbidden_edges: :class:`list`
        "-/-" edges to remove in the expanded graph.
    epsilon: :class:`float`
        Probability to draw a "+" item.

    Returns
    -------
    model: :class:`~stochastic_matching.model.Model`
        Expanded model with 2n nodes that emulates epsilon-coloring.
    """
    n = model.n
    m = model.m
    graph = make_jit_graph(model)
    new_rates = np.concatenate([(1 - epsilon) * model.rates, epsilon * model.rates])
    edge_codex = list()
    for e in range(m):
        nodes = graph.nodes(e)
        for i, offset in enumerate(itertools.product([0, n], repeat=len(nodes))):
            if i == 0 and e in forbidden_edges:
                continue
            edge_codex.append((e, [v + o for v, o in zip(nodes, offset)]))
    new_inc = np.zeros((2 * n, len(edge_codex)), dtype=int)
    for i, nodes in enumerate(edge_codex):
        for node in nodes[1]:
            new_inc[node, i] = 1

    return Model(incidence=new_inc, rates=new_rates), np.array(
        [e[0] for e in edge_codex]
    ).astype(np.int64)


class EFiltering(Simulator):
    """
    Epsilon-filtering policy where incoming items are tagged with a spin.
    A match on a forbidden edge requires at least one "+" item.
    In practice, the simulator works on an expanded graph with twice the initial nodes to represent "+" and "-".

    Parameters
    ----------
    model: :class:`~stochastic_matching.model.Model`
        Model to simulate.
    base_policy: :class:`str` or :class:`~stochastic_matching.simulator.simulator.Simulator`
        Type of simulator to instantiate. Cannot have mandatory extra-parameters (e.g. NOT 'priority').
    forbidden_edges: :class:`list` or :class:`~numpy.ndarray`, optional
        Edges that should not be used.
    weights: :class:`~numpy.ndarray`, optional
        Target rewards on edges. If weights are given, the forbidden edges are computed to match the target
        (overrides forbidden_edges argument).
    epsilon: :class:`float`, default=.01
        Proportion of "+" arrivals (lower means less odds to select a forbidden edge).
    **base_policy_kwargs:
        Keyword arguments. Only universal parameters should be used (seed, max_queue, n_steps).

    Examples
    --------

    Consider the following diamond graph with injective vertex [1, 2, 3]:

    >>> import stochastic_matching as sm
    >>> diamond = sm.CycleChain(rates=[1, 2, 2, 1])


    Without any specific argument, epsilon-filtering acts as a regular longest policy:

    >>> sim = EFiltering(diamond, n_steps=1000, seed=42)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [162 342 348 148]
    Traffic: [ 76  86 190  76  72]
    Queues: [[882  91  23 ...   0   0   0]
     [661 109  72 ...   0   0   0]
     [737 111  63 ...   0   0   0]
     [870  96  27 ...   0   0   0]]
    Steps done: 1000

    Let us use epsilon-filtering by specifying forbidden_edges:

    >>> sim = EFiltering(diamond, forbidden_edges=[0, 4], n_steps=1000, seed=42)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [162 342 348 148]
    Traffic: [  1 158 189 147   1]
    Queues: [[571  44  61 ...   0   0   0]
     [560  37  52 ...   0   0   0]
     [529  66  67 ...   0   0   0]
     [537  62  59 ...   0   0   0]]
    Steps done: 1000

    Switch to a FCFM policy:

    >>> sim = EFiltering(diamond, base_policy='fcfm', forbidden_edges=[0, 4], n_steps=1000, seed=42)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [162 342 348 148]
    Traffic: [  1 158 189 147   1]
    Queues: [[569  99 118 ...   0   0   0]
     [569  35  37 ...   0   0   0]
     [528  58  57 ...   0   0   0]
     [551  84  98 ...   0   0   0]]
    Steps done: 1000

    Switch to virtual queue:

    >>> sim = EFiltering(diamond, base_policy='virtual_queue', forbidden_edges=[0, 4], n_steps=1000, seed=42)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [162 342 348 148]
    Traffic: [  0 157 186 144   1]
    Queues: [[311  93 110 ...   0   0   0]
     [108 120 111 ...   0   0   0]
     [ 67 113  76 ...   0   0   0]
     [174 142 137 ...   0   0   0]]
    Steps done: 1000

    Stolyar's example to see the behavior on hypergraph:

    >>> stol = sm.Model(incidence=[[1, 0, 0, 0, 1, 0, 0],
    ...                  [0, 1, 0, 0, 1, 1, 1],
    ...                  [0, 0, 1, 0, 0, 1, 1],
    ...                  [0, 0, 0, 1, 0, 0, 1]], rates=[1.2, 1.5, 2, .8])
    >>> rewards = [-1, -1, 1, 2, 5, 4, 7]

    >>> sim = EFiltering(stol, base_policy='virtual_queue', rewards=rewards, n_steps=1000, epsilon=.0001, seed=42)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [216 282 371 131]
    Traffic: [  0   0 311  76 213   0  55]
    Queues: [[ 30 114 181 ...   0   0   0]
     [ 22  64 117 ...   0   0   0]
     [ 24 590 217 ...   0   0   0]
     [995   5   0 ...   0   0   0]]
    Steps done: 1000
    >>> int(sim.logs.traffic @ rewards)
    1913
    >>> sim.avg_queues
    array([3.577, 4.856, 1.65 , 0.005])

    To compare with, the original EGPD policy:

    >>> sim = sm.VirtualQueue(stol, rewards=rewards, n_steps=1000, seed=42, beta=.01)
    >>> sim.run()
    >>> sim.plogs
    Arrivals: [236 279 342 143]
    Traffic: [  0   0 100   3 139   0 140]
    Queues: [[  2   3   5 ...   0   0   0]
     [844   8   6 ...   0   0   0]
     [  0   1   3 ...   0   0   0]
     [597 230  96 ...   0   0   0]]
    Steps done: 1000
    >>> int(sim.logs.traffic @ rewards)
    1781
    >>> sim.avg_queues
    array([54.439,  1.597, 78.51 ,  0.691])
    """

    name = "e_filtering"

    def __init__(
        self,
        model,
        base_policy="longest",
        forbidden_edges=None,
        rewards=None,
        epsilon=0.01,
        **base_policy_kwargs,
    ):
        if rewards is not None:
            rewards = np.array(rewards)
            flow = model.optimize_rates(rewards)
            forbidden_edges = [i for i in range(model.m) if flow[i] == 0]
            self.rewards = rewards
        else:
            self.rewards = np.ones(model.m, dtype=int)
            if forbidden_edges is None:
                forbidden_edges = []
            else:
                self.rewards[forbidden_edges] = -1
        self.base_policy = class_converter(base_policy, Simulator)
        self.base_policy_kwargs = base_policy_kwargs
        self.forbidden_edges = forbidden_edges
        self.epsilon = epsilon
        super().__init__(model, **base_policy_kwargs)

    def set_internal(self):
        expanded_model, edges = expand_model(
            model=self.model, forbidden_edges=self.forbidden_edges, epsilon=self.epsilon
        )
        expanded_simu = self.base_policy(
            model=expanded_model, **self.base_policy_kwargs
        )
        expanded_simu.logs = PhantomLogs(
            n=self.model.n, m=self.model.m, max_queue=self.max_queue, edges=edges
        )
        self.internal = {"simu": expanded_simu, "n_steps": self.n_steps}

    def run(self):
        self.internal["simu"].run()
        self.logs = self.internal["simu"].logs



import numpy as np
from stochastic_matching.common import class_converter


class Metric:
    """
    Blueprint for extracting a metric from the simulator.
    """

    name = None

    @staticmethod
    def get(simu):
        """
        Parameters
        ----------
        simu: :class:`~stochastic_matching.simulator.simulator.Simulator`
            Simulator to use.

        Returns
        -------
        Object
            A metric.
        """
        raise NotImplementedError


class Traffic(Metric):
    """
    Matched edges.
    """

    name = "traffic"

    @staticmethod
    def get(simu):
        return simu.logs.traffic


class StepsDone(Metric):
    """
    Simulation duration.
    """

    name = "steps_done"

    @staticmethod
    def get(simu):
        return simu.logs.steps_done


class Income(Metric):
    """
    Number of arrived items per class.
    """

    name = "income"

    @staticmethod
    def get(simu):
        return simu.logs.income


class Regret(Metric):
    """
    Given the matched items observed, difference between the best possible match
    between these items and the actual match.

    Notes
    -----

    * Regret depends on the assigned rewards.
    * Unmatched items are ignored to avoid side effects (e.g. if the optimal allocation is
      unstable, for example because of negative weights).
    """

    name = "regret"

    @staticmethod
    def get(simu):
        rewards = getattr(simu, "rewards", np.ones(simu.model.m, dtype=int))
        original_rates = simu.model.rates
        flow = Flow.get(simu)
        simu.model.rates = simu.model.incidence @ flow
        best_flow = simu.model.optimize_rates(rewards)
        simu.model.rates = original_rates
        return rewards @ (best_flow - flow)


class AvgQueues(Metric):
    """
    Average queue size per item class.
    """

    name = "avg_queues"

    @staticmethod
    def get(simu):
        return (simu.logs.queue_log @ np.arange(simu.max_queue)) / simu.logs.steps_done


class Delay(Metric):
    """Average waiting time (i.e. average total queue size divided by global arrival rate)."""

    name = "delay"

    @staticmethod
    def get(simu):
        return np.sum(AvgQueues.get(simu)) / np.sum(simu.model.rates)


class CCDF(Metric):
    """
    Complementary Cumulative Distribution Function of queue size for each item class.
    """

    name = "ccdf"

    @staticmethod
    def get(simu):
        steps = simu.logs.steps_done
        n = simu.model.n
        return (
            steps
            - np.cumsum(np.hstack([np.zeros((n, 1)), simu.logs.queue_log]), axis=1)
        ) / steps


class Flow(Metric):
    """
    Edge matching rate.
    """

    name = "flow"

    @staticmethod
    def get(simu):
        tot_mu = np.sum(simu.model.rates)
        steps = simu.logs.steps_done
        return simu.logs.traffic * tot_mu / steps


def get_metrics(simu, metrics):
    """
    Extraction of metrics from a simulator. Metrics can be:

    * Name of a :class:`~stochastic_matching.simulator.metrics.Metric` subclass
    * A :class:`~stochastic_matching.simulator.metrics.Metric` subclass
    * A callable that takes a :class:`~stochastic_matching.simulator.simulator.Simulator` as input
      (the name will be the name of the callable)
    * A list of metrics defined as above

    Parameters
    ----------
    simu: :class:`~stochastic_matching.simulator.simulator.Simulator`
        Simulator to use.
    metrics: :class:`str` or :class:`~stochastic_matching.simulator.metrics.Metric` or :class:`list` or callable.
        Metric(s) to extract.
    Returns
    -------
    :class:`dict`
        Metric(s).

    """
    if isinstance(metrics, list):
        return {
            k: v for metric in metrics for k, v in get_metrics(simu, metric).items()
        }
    try:
        getter = class_converter(metrics, Metric)
        return {getter.name: getter.get(simu)}
    except TypeError:
        return {metrics.__name__: metrics(simu)}
